from __future__ import annotations

import uuid
from enum import Enum
from typing import Any, Dict, Optional, Tuple

from sqlalchemy import Boolean, Column, DateTime
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import ForeignKey, Index, String, UniqueConstraint, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Query, Session, relationship
from sqlalchemy_utils import StringEncryptedType
from sqlalchemy_utils.types.encrypted.encrypted_type import AesGcmEngine

from fides.api.db.base_class import Base, JSONTypeOverride
from fides.api.migrations.hash_migration_mixin import HashMigrationMixin
from fides.api.models.privacy_notice import (
    ConsentMechanism,
    PrivacyNoticeHistory,
    UserConsentPreference,
)
from fides.api.models.privacy_request import PrivacyRequest, ProvidedIdentity
from fides.api.models.worker_task import ExecutionLogStatus
from fides.api.schemas.language import SupportedLanguage
from fides.api.schemas.redis_cache import MultiValue
from fides.config import CONFIG


class RequestOrigin(Enum):
    """
    We currently extract the RequestOrigin from the Privacy Experience
    Config ComponentType when saving, so RequestOrigin needs to be a
    superset of ComponentType.

    Not at the db level due to being subject to change.
    Only add here, do not remove
    """

    privacy_center = "privacy_center"
    overlay = "overlay"  # DEPRECATED. DO NOT REMOVE.
    modal = "modal"
    banner_and_modal = "banner_and_modal"
    api = "api"
    tcf_overlay = "tcf_overlay"


class ConsentMethod(Enum):
    # Not at the db level due to being subject to change.
    # Only add here, do not remove
    button = "button"  # deprecated- keeping for backwards-compatibility
    reject = "reject"
    accept = "accept"
    script = "script"
    save = "save"
    dismiss = "dismiss"
    gpc = "gpc"
    individual_notice = "individual_notice"
    acknowledge = "acknowledge"


class ServingComponent(Enum):
    """
    This differs from component type because we want to record exactly
    where consent was served.

    Not at the db level due to being subject to change.
    Only add here, do not remove
    """

    overlay = "overlay"  # DEPRECATED. DO NOT REMOVE.
    banner = "banner"
    modal = "modal"
    privacy_center = "privacy_center"
    tcf_overlay = "tcf_overlay"  # TCF modal in this case
    tcf_banner = "tcf_banner"


class ConsentIdentitiesMixin(HashMigrationMixin):
    """Encrypted and hashed identities for consent reporting and last saved preference retrieval"""

    email = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )  # Encrypted email

    fides_user_device = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )  # Encrypted fides user device

    phone_number = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )  # Encrypted phone number

    external_id = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )

    hashed_email = Column(
        String,
        index=True,
    )  # For exact match searches

    hashed_fides_user_device = Column(String, index=True)  # For exact match searches

    hashed_phone_number = Column(
        String,
        index=True,
    )  # For exact match searches

    hashed_external_id = Column(
        String,
        index=True,
    )  # For exact match searches

    @classmethod
    def bcrypt_hash_value(
        cls,
        value: MultiValue,
        encoding: str = "UTF-8",
    ) -> Optional[str]:
        """
        Temporary function used to hash values to the previously used bcrypt hashes.
        This can be removed once the bcrypt to SHA-256 migration is complete.
        """
        if not value:
            return None

        return ProvidedIdentity.bcrypt_hash_value(value, encoding)

    @classmethod
    def hash_value(
        cls,
        value: Optional[MultiValue] = None,
        encoding: str = "UTF-8",
    ) -> Optional[str]:
        """Utility function to hash the value with a generated salt
        This returns None if there's no value, unlike ProvidedIdentity.hash_value
        """
        if not value:
            return None

        return ProvidedIdentity.hash_value(value, encoding)

    def migrate_hashed_fields(self) -> None:
        if unencrypted_email := self.email:
            self.hashed_email = self.hash_value(unencrypted_email)
        if unencrypted_fides_user_device := self.fides_user_device:
            self.hashed_fides_user_device = self.hash_value(
                unencrypted_fides_user_device
            )
        if unencrypted_phone_number := self.phone_number:
            self.hashed_phone_number = self.hash_value(unencrypted_phone_number)
        if unencrypted_external_id := self.external_id:
            self.hashed_external_id = self.hash_value(unencrypted_external_id)
        self.is_hash_migrated = True


class CurrentPrivacyPreference(ConsentIdentitiesMixin, Base):
    """Stores the latest saved privacy preferences for a given user

    Email/phone/fides device must be unique.  If we later tie identities together, these records
    are consolidated.
    """

    @declared_attr
    def __tablename__(self) -> str:
        """Table name is currentprivacypreferencev2 - currentprivacypreference was deprecated and removed"""
        return "currentprivacypreferencev2"

    created_at = Column(DateTime(timezone=True), server_default=func.now(), index=True)

    preferences = Column(
        MutableDict.as_mutable(JSONB)
    )  # All saved preferences against Privacy Notices and TCF Notices

    fides_string = Column(String)

    property_id = Column(
        String,
        nullable=True,
    )

    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        index=True,
    )

    @declared_attr
    def __table_args__(cls: Any) -> Tuple:
        return (
            UniqueConstraint(
                "email",
                "property_id",
                name="last_saved_for_email_per_property_id",
            ),
            UniqueConstraint(
                "phone_number",
                "property_id",
                name="last_saved_for_phone_number_per_property_id",
            ),
            UniqueConstraint(
                "fides_user_device",
                "property_id",
                name="last_saved_for_fides_user_device_per_property_id",
            ),
            UniqueConstraint(
                "external_id",
                "property_id",
                name="last_saved_for_external_id_per_property_id",
            ),
            Index(
                "idx_currentprivacypreferencev2_unmigrated",
                "is_hash_migrated",
                postgresql_where=cls.is_hash_migrated.is_(  # pylint: disable=no-member
                    False
                ),
            ),
        )


class LastServedNotice(Base):
    """
    DEPRECATED. DO NOT UPDATE THIS TABLE.  This will soon be removed.

    This table consolidates every notice a user has been served, (analogous to CurrentPrivacyPreference
    but it is being removed). Backend is not writing to this any longer.
    """

    @declared_attr
    def __tablename__(self) -> str:
        """Table name is lastservednoticev2 - lastservednotice was deprecated and removed"""
        return "lastservednoticev2"

    created_at = Column(DateTime(timezone=True), server_default=func.now(), index=True)

    served = Column(
        MutableDict.as_mutable(JSONB)
    )  # All Privacy Notices and TCF Notices Served

    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        index=True,
    )

    __table_args__ = (
        UniqueConstraint(
            "email",
            name="last_served_for_email",
        ),
        UniqueConstraint(
            "phone_number",
            name="last_served_for_phone_number",
        ),
        UniqueConstraint(
            "fides_user_device",
            name="last_served_for_fides_user_device",
        ),
    )

    email = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )  # Encrypted email

    fides_user_device = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )  # Encrypted fides user device

    phone_number = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )  # Encrypted phone number

    hashed_email = Column(
        String,
        index=True,
    )  # For exact match searches

    hashed_fides_user_device = Column(String, index=True)  # For exact match searches

    hashed_phone_number = Column(
        String,
        index=True,
    )  # For exact match searches

    @classmethod
    def generate_served_notice_history_id(cls) -> str:
        """Generate a served notice history id

        We save privacy preferences and return immediately alongside this generated
        id that we build up front.  This value will be saved under related ServedNoticeHistory.served_notice_history_id
        fields, and then passed in to save privacy preferences and saved under PrivacyPreferenceHistory.served_notice_history_id
        """
        return f"ser_{uuid.uuid4()}"


class ConsentReportingMixinV2(ConsentIdentitiesMixin):
    """Consent Reporting Mixin to share between historical preferences saved and
    historical notices served
    """

    anonymized_ip_address = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )
    created_at = Column(DateTime(timezone=True), server_default=func.now(), index=True)

    language = Column(
        EnumColumn(
            SupportedLanguage,
            native_enum=False,
            values_callable=lambda x: [i.value for i in x],
        ),
        index=True,
    )

    notice_key = Column(String, index=True)  # Privacy Notice Key

    notice_mechanism = Column(
        EnumColumn(ConsentMechanism), index=True
    )  # Privacy Notice Mechanism

    notice_name = Column(String, index=True)  # Privacy Notice name or "TCF"

    property_id = Column(
        String,
        index=True,
        nullable=True,
    )

    # The specific version of the experience config the user was shown to present the relevant notice
    # Contains the version, language, button labels, description, etc.
    @declared_attr
    def privacy_experience_config_history_id(cls) -> Column:
        return Column(
            String,
            ForeignKey("privacyexperienceconfighistory.id"),
            index=True,
        )

    @declared_attr
    def privacy_notice_history_id(cls) -> Column:
        """
        The specific historical record the user consented to - applicable when
        saving preferences with respect to a privacy notice directly
        """
        return Column(String, ForeignKey(PrivacyNoticeHistory.id), index=True)

    # Preferences and Notices Served are saved in celery - there may be some gap in between
    # when the data was received, and when we actually were able to save it to the db
    received_at = Column(DateTime(timezone=True))

    # Location where we received the request
    request_origin = Column(EnumColumn(RequestOrigin))

    url_recorded = Column(String)

    user_agent = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )

    user_geography = Column(String, index=True)

    # Relationships
    @declared_attr
    def privacy_notice_history(cls) -> relationship:
        return relationship(PrivacyNoticeHistory)


class ServedNoticeHistory(ConsentReportingMixinV2, Base):
    """A Historical Record of every time a Notice was served to a user

    Each Privacy Notice served gets its own record, while served TCF attributes are collapsed into one record.

    The "served_notice_history_id" column on this table can be mapped to the PrivacyPreferenceHistory records
    to calculate conversion.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "servednoticehistory"

    acknowledge_mode = Column(
        Boolean,
        default=False,
    )

    serving_component = Column(EnumColumn(ServingComponent), nullable=False, index=True)

    # Generated identifier for the ServedNoticeHistory, used to link a ServedNoticeHistory and PrivacyPreferenceHistory
    # record together
    served_notice_history_id = Column(String, index=True)

    tcf_served = Column(
        MutableDict.as_mutable(JSONB)
    )  # Dict of all the TCF attributes served if applicable

    @staticmethod
    def get_by_served_id(db: Session, served_id: str) -> Query:
        """Retrieves all ServedNoticeHistory records with a common served_notice_history_id - generated
        before the task was queued to store these records
        """
        return db.query(ServedNoticeHistory).filter(
            ServedNoticeHistory.served_notice_history_id == served_id
        )


class PrivacyPreferenceHistory(ConsentReportingMixinV2, Base):
    """A Historical Record of every time a Notice was saved for a user

    Each Privacy Notice served gets its own record, while served TCF attributes are collapsed into one record.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "privacypreferencehistory"

    # Systems capable of propagating their consent, and their status.  If the preference is
    # not relevant for the system, or we couldn't propagate a preference, the status is skipped
    affected_system_status = Column(
        MutableDict.as_mutable(JSONB), server_default="{}", default=dict
    )

    # Optional, if fides_string was supplied directly
    fides_string = Column(String)

    # Accept, reject, etc. especially useful for TCF
    method = Column(EnumColumn(ConsentMethod))

    # Whether the user wants to opt in, opt out, or has acknowledged the notice.
    # For TCF notices, we just say "tcf", and more detailed preferences are under "tcf_preferences"
    preference = Column(EnumColumn(UserConsentPreference), index=True, nullable=False)

    # The privacy request created to propagate the preferences if applicable.  The notice
    # must have had system wide enforcement. TCF notices don't have privacy requests created.
    privacy_request_id = Column(
        String, ForeignKey(PrivacyRequest.id, ondelete="SET NULL"), index=True
    )

    # Relevant identities are added to the report during request propagation
    secondary_user_ids = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
    )

    # The record of where we served the notice in the frontend, for conversion purposes.
    served_notice_history_id = Column(String, index=True)

    tcf_preferences = Column(
        MutableDict.as_mutable(JSONB)
    )  # Dict of TCF attributes saved, for a TCF notice

    privacy_request = relationship(PrivacyRequest, backref="privacy_preferences")

    def cache_system_status(
        self, db: Session, system: str, status: ExecutionLogStatus
    ) -> None:
        """Update the cached affected system status for consent reporting

        Typically this should just be called for consent connectors only.
        If no request is made or email is sent, this should be called with a status of skipped.
        """
        if not self.affected_system_status:
            self.affected_system_status = {}
        self.affected_system_status[system] = (
            status.name
        )  # To avoid using "ExecutionLogStatus.paused" in the logs
        self.save(db)

    def update_secondary_user_ids(
        self, db: Session, new_identities: Dict[str, Any]
    ) -> None:
        """Update secondary user identities for consent reporting

        The intent is to only put identities here that we intend to send to third party systems.
        """
        secondary_user_ids = self.secondary_user_ids or {}
        secondary_user_ids.update(new_identities)
        self.secondary_user_ids = secondary_user_ids
        self.save(db)



"""add version fields to default types

Revision ID: 708a780b01ba
Revises: 192f23f4c968
Create Date: 2023-08-17 12:29:04.855626

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "708a780b01ba"
down_revision = "192f23f4c968"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind: Connection = op.get_bind()
    bind.execute(sa.text("DELETE FROM ctl_data_uses WHERE is_default = TRUE;"))
    bind.execute(sa.text("DELETE FROM ctl_data_categories WHERE is_default = TRUE;"))
    bind.execute(sa.text("DELETE FROM ctl_data_qualifiers WHERE is_default = TRUE;"))
    bind.execute(sa.text("DELETE FROM ctl_data_subjects WHERE is_default = TRUE;"))

    op.add_column(
        "ctl_data_categories", sa.Column("version_added", sa.Text(), nullable=True)
    )
    op.add_column(
        "ctl_data_categories", sa.Column("version_deprecated", sa.Text(), nullable=True)
    )
    op.add_column(
        "ctl_data_categories", sa.Column("replaced_by", sa.Text(), nullable=True)
    )
    op.add_column(
        "ctl_data_qualifiers", sa.Column("version_added", sa.Text(), nullable=True)
    )
    op.add_column(
        "ctl_data_qualifiers", sa.Column("version_deprecated", sa.Text(), nullable=True)
    )
    op.add_column(
        "ctl_data_qualifiers", sa.Column("replaced_by", sa.Text(), nullable=True)
    )
    op.add_column(
        "ctl_data_subjects", sa.Column("version_added", sa.Text(), nullable=True)
    )
    op.add_column(
        "ctl_data_subjects", sa.Column("version_deprecated", sa.Text(), nullable=True)
    )
    op.add_column(
        "ctl_data_subjects", sa.Column("replaced_by", sa.Text(), nullable=True)
    )
    op.add_column("ctl_data_uses", sa.Column("version_added", sa.Text(), nullable=True))
    op.add_column(
        "ctl_data_uses", sa.Column("version_deprecated", sa.Text(), nullable=True)
    )
    op.add_column("ctl_data_uses", sa.Column("replaced_by", sa.Text(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("ctl_data_uses", "replaced_by")
    op.drop_column("ctl_data_uses", "version_deprecated")
    op.drop_column("ctl_data_uses", "version_added")
    op.drop_column("ctl_data_subjects", "replaced_by")
    op.drop_column("ctl_data_subjects", "version_deprecated")
    op.drop_column("ctl_data_subjects", "version_added")
    op.drop_column("ctl_data_qualifiers", "replaced_by")
    op.drop_column("ctl_data_qualifiers", "version_deprecated")
    op.drop_column("ctl_data_qualifiers", "version_added")
    op.drop_column("ctl_data_categories", "replaced_by")
    op.drop_column("ctl_data_categories", "version_deprecated")
    op.drop_column("ctl_data_categories", "version_added")
    # ### end Alembic commands ###



from typing import Any, Optional

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.models.attachment import Attachment
from fides.api.models.manual_tasks.manual_task_config import ManualTaskConfigField
from fides.api.models.manual_tasks.manual_task_instance import (
    ManualTaskInstance,
    ManualTaskSubmission,
)
from fides.api.schemas.manual_tasks.manual_task_status import (
    StatusTransitionNotAllowed,
    StatusType,
)
from fides.service.manual_tasks.utils import validate_fields, with_task_logging


class ManualTaskInstanceError(Exception):
    """Exception raised when a manual task instance error occurs."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class ManualTaskSubmissionError(Exception):
    """Exception raised when a manual task submission error occurs."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class ManualTaskInstanceService:
    def __init__(self, db: Session):
        self.db = db

    def _create_log_data(
        self,
        task_id: str,
        config_id: str,
        instance_id: str,
        details: dict[str, Any],
    ) -> dict[str, Any]:
        """Create standard log data structure."""
        return {
            "task_id": task_id,
            "config_id": config_id,
            "instance_id": instance_id,
            "details": details,
        }

    def _get_instance(
        self, instance_id: str, allow_completed: bool = False
    ) -> ManualTaskInstance:
        """Get and validate instance."""
        instance = self.db.query(ManualTaskInstance).filter_by(id=instance_id).first()
        if not instance:
            raise ManualTaskInstanceError(f"Instance with ID {instance_id} not found")

        if not allow_completed and instance.status == StatusType.completed:
            raise StatusTransitionNotAllowed(
                "Instance is already completed, no further changes allowed"
            )

        return instance

    def _get_field(
        self, field_id: str, validate_data: Optional[dict[str, Any]] = None
    ) -> ManualTaskConfigField:
        """Get and validate field."""
        field = ManualTaskConfigField.get_by_key_or_id(
            db=self.db, data={"id": field_id}
        )
        if not field:
            raise ManualTaskInstanceError(f"Field with ID {field_id} not found")

        if validate_data:
            try:
                validate_fields([validate_data], is_submission=True)
            except ValueError as e:
                raise ManualTaskInstanceError(f"Invalid field data: {e}") from e

        return field

    def _update_instance_status(
        self,
        instance: ManualTaskInstance,
        new_status: Optional[StatusType] = None,
        user_id: Optional[str] = None,
        silent: bool = False,
    ) -> None:
        """Update instance status with optional error suppression."""
        try:
            if not new_status:
                new_status = (
                    StatusType.in_progress
                    if instance.submissions
                    else StatusType.pending
                )
            instance.update_status(self.db, new_status, user_id)
        except StatusTransitionNotAllowed as e:
            if not silent:
                raise
            logger.info(f"Status not transitioning: {e}")

    @with_task_logging("Created task instance")
    def create_instance(
        self, task_id: str, config_id: str, entity_id: str, entity_type: str
    ) -> tuple[ManualTaskInstance, dict[str, Any]]:
        """Create a new instance for an entity."""
        instance = ManualTaskInstance.create(
            self.db,
            data={
                "task_id": task_id,
                "config_id": config_id,
                "entity_id": entity_id,
                "entity_type": entity_type,
            },
        )
        return instance, self._create_log_data(
            task_id, config_id, instance.id, {"entity_type": entity_type}
        )

    def get_submission_for_field(
        self, instance_id: str, field_id: str
    ) -> Optional[ManualTaskSubmission]:
        """Get the submission for a specific field."""
        return self._get_instance(instance_id).get_submission_for_field(field_id)

    @with_task_logging("Updated task instance status")
    def update_status(
        self,
        instance_id: str,
        new_status: Optional[StatusType] = None,
        user_id: Optional[str] = None,
    ) -> tuple[ManualTaskInstance, dict[str, Any]]:
        """Update instance status with logging."""
        instance = self._get_instance(instance_id)
        previous_status = instance.status
        self._update_instance_status(instance, new_status, user_id)

        return instance, self._create_log_data(
            instance.task_id,
            instance.config_id,
            instance.id,
            {
                "previous_status": previous_status,
                "new_status": instance.status,
                "user_id": user_id,
            },
        )

    @with_task_logging("Created task submission")
    def create_submission(
        self,
        instance_id: str,
        field_id: str,
        data: dict[str, Any],
    ) -> tuple[ManualTaskSubmission, dict[str, Any]]:
        """Create a new submission for a field."""
        instance = self._get_instance(instance_id)
        field = self._get_field(field_id, data)

        if instance.get_submission_for_field(field_id):
            raise ManualTaskInstanceError(
                f"Submission for field {field.field_key} already exists for instance {instance.id}"
            )

        submission = ManualTaskSubmission.create(
            self.db,
            data={
                "task_id": instance.task_id,
                "config_id": instance.config_id,
                "instance_id": instance.id,
                "field_id": field.id,
                "data": data,
            },
        )

        # Update instance status to in_progress
        self._update_instance_status(instance, StatusType.in_progress, silent=True)

        return submission, self._create_log_data(
            instance.task_id,
            instance.config_id,
            instance.id,
            {
                "field_key": field.field_key,
                "field_type": field.field_type,
            },
        )

    @with_task_logging("Updated task submission")
    def update_submission(
        self,
        instance_id: str,
        submission_id: str,
        data: dict[str, Any],
    ) -> tuple[ManualTaskSubmission, dict[str, Any]]:
        """Update a submission for a field."""
        instance = self._get_instance(instance_id)
        submission = next(
            (s for s in instance.submissions if s.id == submission_id),
            None,
        )
        if not submission or not submission.field_id:
            raise ManualTaskSubmissionError(
                f"Valid submission with ID {submission_id} not found"
            )

        field = self._get_field(submission.field_id, data)
        submission.update(self.db, data={"data": data})
        self._update_instance_status(instance, silent=True)

        return submission, self._create_log_data(
            instance.task_id,
            instance.config_id,
            instance.id,
            {
                "field_key": field.field_key,
                "field_type": field.field_type,
            },
        )

    @with_task_logging("Delete task attachment")
    def delete_attachment_by_id(
        self,
        submission_id: str,
        attachment_id: str,
    ) -> None:
        """Delete an attachment for a field."""
        submission = (
            self.db.query(ManualTaskSubmission).filter_by(id=submission_id).first()
        )
        if not submission:
            raise ManualTaskSubmissionError(
                f"Submission with ID {submission_id} does not exist"
            )

        self._get_instance(
            submission.instance_id
        )  # Validates instance is not completed

        attachment = self.db.query(Attachment).filter_by(id=attachment_id).first()
        if not attachment or attachment not in submission.attachments:
            raise ManualTaskSubmissionError(
                f"Attachment {attachment_id} not found in submission {submission_id}"
            )

        # Delete attachment and optionally submission
        if (
            len(submission.attachments) == 1
            and submission.field.field_type == "attachment"
        ):
            attachment.delete(self.db)
            submission.delete(self.db)
        else:
            attachment.delete(self.db)

    @with_task_logging("Completed task instance")
    def complete_task_instance(
        self, task_id: str, config_id: str, instance_id: str, user_id: str
    ) -> tuple[ManualTaskInstance, dict[str, Any]]:
        """Complete a task instance."""
        instance = self._get_instance(instance_id)

        missing_fields = [
            field.field_key
            for field in instance.required_fields
            if not instance.get_submission_for_field(field.id)
        ]
        if missing_fields:
            raise StatusTransitionNotAllowed(
                f"Cannot complete task instance. Missing required fields: {', '.join(missing_fields)}"
            )

        instance.update_status(self.db, StatusType.completed, user_id)
        instance.completed_by_id = user_id
        instance.save(self.db)

        return instance, self._create_log_data(
            task_id, config_id, instance_id, {"completed_by": user_id}
        )



"""add_monitor_tasks

Revision ID: 29e56fa1fdb3
Revises: 5efcdf18438e
Create Date: 2025-06-11 14:40:08.384571

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "29e56fa1fdb3"
down_revision = "5efcdf18438e"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "monitortask",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("action_type", sa.String(), nullable=False),
        sa.Column(
            "status",
            sa.Enum(
                "in_processing",
                "pending",
                "complete",
                "error",
                "paused",
                "retrying",
                "skipped",
                name="executionlogstatus",
                native_enum=False,
            ),
            nullable=False,
        ),
        sa.Column("celery_id", sa.String(length=255), nullable=False),
        sa.Column(
            "task_arguments", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.Column("message", sa.String(), nullable=True),
        sa.Column("monitor_config_id", sa.String(), nullable=False),
        sa.Column("staged_resource_urns", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("child_resource_urns", sa.ARRAY(sa.String()), nullable=True),
        sa.ForeignKeyConstraint(
            ["monitor_config_id"], ["monitorconfig.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("celery_id"),
    )
    op.create_index(
        op.f("ix_monitortask_action_type"), "monitortask", ["action_type"], unique=False
    )
    op.create_index(op.f("ix_monitortask_id"), "monitortask", ["id"], unique=False)
    op.create_index(
        op.f("ix_monitortask_monitor_config_id"),
        "monitortask",
        ["monitor_config_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_monitortask_status"), "monitortask", ["status"], unique=False
    )
    op.create_table(
        "monitortaskexecutionlog",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "status",
            postgresql.ENUM(name="executionlogstatus", create_type=False),
            nullable=False,
        ),
        sa.Column("message", sa.String(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("clock_timestamp()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("clock_timestamp()"),
            nullable=True,
        ),
        sa.Column("celery_id", sa.String(length=255), nullable=False),
        sa.Column("monitor_task_id", sa.String(), nullable=False),
        sa.Column(
            "run_type", sa.Enum("MANUAL", "SYSTEM", name="taskruntype"), nullable=False
        ),
        sa.ForeignKeyConstraint(
            ["monitor_task_id"], ["monitortask.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_monitortaskexecutionlog_id"),
        "monitortaskexecutionlog",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_monitortaskexecutionlog_monitor_task_id"),
        "monitortaskexecutionlog",
        ["monitor_task_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_monitortaskexecutionlog_status"),
        "monitortaskexecutionlog",
        ["status"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_monitortaskexecutionlog_status"), table_name="monitortaskexecutionlog"
    )
    op.drop_index(
        op.f("ix_monitortaskexecutionlog_monitor_task_id"),
        table_name="monitortaskexecutionlog",
    )
    op.drop_index(
        op.f("ix_monitortaskexecutionlog_id"), table_name="monitortaskexecutionlog"
    )
    op.drop_table("monitortaskexecutionlog")
    op.drop_index(op.f("ix_monitortask_status"), table_name="monitortask")
    op.drop_index(op.f("ix_monitortask_monitor_config_id"), table_name="monitortask")
    op.drop_index(op.f("ix_monitortask_id"), table_name="monitortask")
    op.drop_index(op.f("ix_monitortask_action_type"), table_name="monitortask")
    op.drop_table("monitortask")
    op.execute("DROP TYPE IF EXISTS taskruntype")



from typing import Any, ContextManager, Dict, List, Optional

from celery import Celery, Task
from loguru import logger
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session
from tenacity import (
    RetryCallState,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from fides.api.db.session import get_db_engine, get_db_session
from fides.api.util.logger import setup as setup_logging
from fides.config import CONFIG, FidesConfig

MESSAGING_QUEUE_NAME = "fidesops.messaging"
PRIVACY_PREFERENCES_QUEUE_NAME = "fides.privacy_preferences"  # This queue is used in Fidesplus for saving privacy preferences and notices served
DSR_QUEUE_NAME = "fides.dsr"  # This queue is used for running data subject requests
DISCOVERY_MONITORS_DETECTION_QUEUE_NAME = "fidesplus.discovery_monitors_detection"  # This queue is used for running discovery monitors detection tasks
DISCOVERY_MONITORS_CLASSIFICATION_QUEUE_NAME = "fidesplus.discovery_monitors_classification"  # This queue is used for running discovery monitors classification tasks
DISCOVERY_MONITORS_PROMOTION_QUEUE_NAME = "fidesplus.discovery_monitors_promotion"  # This queue is used for running discovery monitors promotion tasks


NEW_SESSION_RETRIES = 5

autodiscover_task_locations: List[str] = [
    "fides.api.tasks",
    "fides.api.tasks.scheduled",
    "fides.api.service.privacy_request",
    "fides.api.service.privacy_request.request_runner_service",
]


def log_retry_attempt(retry_state: RetryCallState) -> None:
    """Log database connection retry attempts."""

    logger.warning(
        "Database connection attempt {} failed. Retrying in {} seconds...",
        retry_state.attempt_number,
        retry_state.next_action.sleep,  # type: ignore[union-attr]
    )


class DatabaseTask(Task):  # pylint: disable=W0223
    _task_engine = None
    _sessionmaker = None

    # This retry will attempt to connect 5 times with an exponential backoff (2, 4, 8, 16 seconds between each attempt).
    # The original error will be re-raised if the retries are successful. All attempts are shown in the logs.
    @retry(
        stop=stop_after_attempt(NEW_SESSION_RETRIES),
        wait=wait_exponential(multiplier=1, min=1),
        retry=retry_if_exception_type(OperationalError),
        reraise=True,
        before_sleep=log_retry_attempt,
    )
    def get_new_session(self) -> ContextManager[Session]:
        """
        Creates a new Session to be used for each task invocation.

        The new Sessions will reuse a shared `Engine` and `sessionmaker`
        across invocations, so as to reuse db connection resources.
        """
        # only one engine will be instantiated in a given task scope, i.e
        # once per celery process.
        if self._task_engine is None:
            self._task_engine = get_db_engine(
                config=CONFIG,
                pool_size=CONFIG.database.task_engine_pool_size,
                max_overflow=CONFIG.database.task_engine_max_overflow,
                keepalives_idle=CONFIG.database.task_engine_keepalives_idle,
                keepalives_interval=CONFIG.database.task_engine_keepalives_interval,
                keepalives_count=CONFIG.database.task_engine_keepalives_count,
            )

        # same for the sessionmaker
        if self._sessionmaker is None:
            self._sessionmaker = get_db_session(config=CONFIG, engine=self._task_engine)

        # but a new session is instantiated each time the method is invoked
        # to prevent session overlap when requests are executing concurrently
        # when in task_always_eager mode (i.e. without proper workers)
        new_session = self._sessionmaker()
        logger.debug(f"DatabaseTaskSession ID: {id(new_session)}. Self ID: {id(self)}")
        return new_session


def _create_celery(config: FidesConfig = CONFIG) -> Celery:
    """
    Returns a configured version of the Celery application
    """
    setup_logging(config)
    logger.bind(api_config=CONFIG.logging.json()).debug(
        "Logger configuration options in use"
    )

    app = Celery(__name__)

    celery_config: Dict[str, Any] = {
        # Defaults for the celery config
        "broker_url": config.redis.connection_url,
        "result_backend": config.redis.connection_url,
        "event_queue_prefix": "fides_worker",
        "task_always_eager": True,
        # Ops requires this to route emails to separate queues
        "task_create_missing_queues": True,
        "task_default_queue": "fides",
    }

    celery_config.update(config.celery)

    app.conf.update(celery_config)

    app.autodiscover_tasks(autodiscover_task_locations)

    return app


celery_app = _create_celery(CONFIG)


def get_worker_ids() -> List[Optional[str]]:
    """
    Returns a list of the connected healthy worker UUIDs.
    """
    try:
        connected_workers = [
            key for key, _ in celery_app.control.inspect().ping().items()
        ]
    except Exception as exception:
        logger.critical(exception)
        connected_workers = []
    return connected_workers



from typing import List, cast

from sqlalchemy import ARRAY, Column, ForeignKey, String
from sqlalchemy.orm import Session, relationship

from fides.api.db.base_class import Base
from fides.api.models.fides_user import FidesUser
from fides.api.oauth.roles import (
    EXTERNAL_RESPONDENT,
    RESPONDENT,
    ROLES_TO_SCOPES_MAPPING,
)


class FidesUserPermissions(Base):
    """The DB ORM model for FidesUserPermissions"""

    user_id = Column(String, ForeignKey(FidesUser.id), nullable=False, unique=True)
    roles = Column(ARRAY(String), nullable=False, server_default="{}", default=dict)
    user = relationship(FidesUser, back_populates="permissions", uselist=False)

    @property
    def total_scopes(self) -> List[str]:
        """Returns the scopes the user has inherited via their roles."""
        all_scopes = []
        for role in self.roles:
            all_scopes += ROLES_TO_SCOPES_MAPPING.get(role, [])

        return sorted(list(set(all_scopes)))

    def is_respondent(self) -> bool:
        """Check if the user is a respondent."""
        return any(role in self.roles for role in [RESPONDENT, EXTERNAL_RESPONDENT])

    def update_roles(self, db: Session, new_roles: List[str]) -> None:
        """Update the user's roles if allowed.
        Raises ValueError if role changes are not allowed."""
        if self.is_respondent():
            raise ValueError("Role changes are not allowed for respondents")

        self.roles = new_roles
        self.save(db)

    def update(self, db: Session, *, data: dict) -> "FidesUserPermissions":
        """Update the user permissions with the provided data."""
        return cast(FidesUserPermissions, super().update(db, data=data))



"""add `data_uses` column to stagedresource

Revision ID: d9237a0c0d5a
Revises: ae65da77c468
Create Date: 2024-11-21 13:18:24.085858

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d9237a0c0d5a"
down_revision = "ae65da77c468"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "stagedresource", sa.Column("data_uses", sa.ARRAY(sa.String), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    op.drop_column("stagedresource", "data_uses")
    # ### end Alembic commands ###



from datetime import datetime, timezone
from enum import Enum as EnumType
from typing import Optional, Protocol

from sqlalchemy.orm import Session


class StatusTransitionNotAllowed(Exception):
    """Exception raised when a status transition is not allowed."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class StatusType(str, EnumType):
    """Enum for manual task status."""

    pending = "pending"
    in_progress = "in_progress"
    completed = "completed"
    failed = "failed"

    @classmethod
    def get_valid_transitions(cls, current_status: "StatusType") -> list["StatusType"]:
        """Get valid transitions from the current status."""
        transitions = {
            cls.pending: [cls.in_progress, cls.failed, cls.completed],
            cls.in_progress: [cls.completed, cls.failed],
            cls.completed: [],
            cls.failed: [cls.pending, cls.in_progress],
        }
        return transitions.get(current_status, [])


class StatusTransitionProtocol(Protocol):
    """Protocol for objects that support status transitions.

    This protocol defines the interface that any object supporting status transitions
    must implement. It includes both the required attributes and methods.

    Example:
        ```python
        # Any class that implements this protocol can be used interchangeably
        def process_status_update(obj: StatusTransitionProtocol, db: Session) -> None:
            if obj.is_pending:
                obj.start_progress(db)
            elif obj.is_in_progress:
                obj.mark_completed(db, user_id="user123")

        # This works with ManualTaskInstance or any other class implementing the protocol
        instance = ManualTaskInstance(...)
        process_status_update(instance, db)
        ```
    """

    # Required attributes - using runtime types that work with SQLAlchemy
    status: StatusType
    completed_at: Optional[datetime]  # Can be None when resetting to pending
    completed_by_id: Optional[str]  # Can be None when resetting to pending

    # Required methods
    # pylint does not understand the Protocol abstract syntax and will complain about the ellipsis
    def update_status(
        self, db: Session, new_status: StatusType, user_id: Optional[str] = None
    ) -> None:
        """Update the status with validation and completion handling."""
        ...  # pylint: disable=unnecessary-ellipsis

    def mark_completed(self, db: Session, user_id: str) -> None:
        """Mark as completed."""
        ...  # pylint: disable=unnecessary-ellipsis

    def mark_failed(self, db: Session) -> None:
        """Mark as failed."""
        ...  # pylint: disable=unnecessary-ellipsis

    def start_progress(self, db: Session) -> None:
        """Mark as in progress."""
        ...  # pylint: disable=unnecessary-ellipsis

    def reset_to_pending(self, db: Session) -> None:
        """Reset to pending status."""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def is_completed(self) -> bool:
        """Check if completed."""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def is_failed(self) -> bool:
        """Check if failed."""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def is_in_progress(self) -> bool:
        """Check if in progress."""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def is_pending(self) -> bool:
        """Check if pending."""
        ...  # pylint: disable=unnecessary-ellipsis


def validate_status_transition_object(obj: StatusTransitionProtocol) -> bool:
    """Validate that an object properly implements the StatusTransitionProtocol.

    This function demonstrates how the Protocol can be used for runtime validation
    and type checking.
    """
    required_attrs = ["status", "completed_at", "completed_by_id"]
    required_methods = [
        "update_status",
        "mark_completed",
        "mark_failed",
        "start_progress",
        "reset_to_pending",
    ]
    required_properties = ["is_completed", "is_failed", "is_in_progress", "is_pending"]

    # Check all required elements
    all_required = required_attrs + required_methods + required_properties
    return all(hasattr(obj, attr) for attr in all_required) and all(
        callable(getattr(obj, method)) for method in required_methods
    )


class StatusTransitionMixin:
    """Mixin for handling status transitions.

    This mixin provides methods for managing status transitions and completion tracking.
    It implements the StatusTransitionProtocol and can be used by any model that needs status management.
    """

    # Type annotations to match the Protocol
    status: StatusType
    completed_at: Optional[datetime]
    completed_by_id: Optional[str]

    def _get_valid_transitions(self) -> list[StatusType]:
        """Get valid transitions from the current status."""
        return StatusType.get_valid_transitions(self.status)

    def _validate_status_transition(self, new_status: StatusType) -> None:
        """Validate that a status transition is allowed."""
        if new_status == self.status:
            raise StatusTransitionNotAllowed(
                f"Invalid status transition: already in status {new_status}"
            )

        valid_transitions = self._get_valid_transitions()
        if new_status not in valid_transitions:
            raise StatusTransitionNotAllowed(
                f"Invalid status transition from {self.status} to {new_status}. "
                f"Valid transitions are: {valid_transitions}"
            )

    def update_status(
        self, db: Session, new_status: StatusType, user_id: Optional[str] = None
    ) -> None:
        """Update the status with validation and completion handling."""
        self._validate_status_transition(new_status)

        if new_status == StatusType.completed:
            self.completed_at = datetime.now(timezone.utc)
            self.completed_by_id = user_id
        elif new_status == StatusType.pending:
            # Reset completion fields if going back to pending
            self.completed_at = None
            self.completed_by_id = None

        self.status = new_status
        db.add(self)
        db.commit()

    def mark_completed(self, db: Session, user_id: str) -> None:
        """Mark as completed."""
        self.update_status(db, StatusType.completed, user_id)

    def mark_failed(self, db: Session) -> None:
        """Mark as failed."""
        self.update_status(db, StatusType.failed)

    def start_progress(self, db: Session) -> None:
        """Mark as in progress."""
        self.update_status(db, StatusType.in_progress)

    def reset_to_pending(self, db: Session) -> None:
        """Reset to pending status."""
        self.update_status(db, StatusType.pending)

    @property
    def is_completed(self) -> bool:
        """Check if completed."""
        return self.status == StatusType.completed

    @property
    def is_failed(self) -> bool:
        """Check if failed."""
        return self.status == StatusType.failed

    @property
    def is_in_progress(self) -> bool:
        """Check if in progress."""
        return self.status == StatusType.in_progress

    @property
    def is_pending(self) -> bool:
        """Check if pending."""
        return self.status == StatusType.pending



from typing import Any, Dict

from pydantic import BaseModel


class BulkResponse(BaseModel):
    """
    Schema for responses from bulk update/create requests.  Override to set "succeeded" and "failed" attributes on
    your child class with paginated types.

    Example:
    from fastapi_pagination import Page

    succeeded: List[ConnectionConfigurationResponse]
    failed: List[BulkUpdateFailed]
    """

    @classmethod
    def __pydantic_init_subclass__(cls: BaseModel, **kwargs):  # type: ignore
        super().__pydantic_init_subclass__(**kwargs)  # type: ignore
        if "succeeded" not in cls.model_fields or "failed" not in cls.model_fields:
            raise TypeError(
                f"Class {cls.__name__} needs both 'succeeded' and 'failed' attributes defined."  # type: ignore
            )


class BulkUpdateFailed(BaseModel):
    """Schema for use when Bulk Create/Update fails."""

    message: str
    data: Dict[str, Any]



import json
from typing import Any, Dict, List

from fastapi import Request
from loguru import logger
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session
from starlette.types import Message

from fides.api.api import deps
from fides.api.models.sql_models import AuditLogResource  # type: ignore[attr-defined]
from fides.api.oauth.utils import extract_token_and_load_client


async def handle_audit_log_resource(request: Request) -> None:
    """
    Handles the lifecycle of recording audit log resource data.

    Attempts to track the WHO, WHEN, and beginning of WHAT for
    traceability purposes.

    WHO: User ID from the API request
    WHEN: Timestamps related to the request
    WHAT: The endpoint, request type, and (if applicable)
    fides_key(s) associated with the request
    """

    # details to be stored as a row on the server
    audit_log_resource_data = {
        "user_id": None,
        "request_path": request.scope["path"],
        "request_type": request.method,
        "fides_keys": None,
        "extra_data": None,
    }
    db: Session = deps.get_api_session()

    # get the user id associated with the request
    token = request.headers.get("authorization")
    if token:
        client = await get_client_user_id(db, token)
        audit_log_resource_data["user_id"] = client

    # Access request body to check for fides_keys
    await set_body(request, await request.body())

    body = await get_body(request)
    fides_keys: List = await extract_data_from_body(body)
    audit_log_resource_data["fides_keys"] = fides_keys

    # write record to server
    await write_audit_log_resource_record(db, audit_log_resource_data)


async def write_audit_log_resource_record(
    db: Session, audit_log_resource_data: Dict[str, Any]
) -> None:
    """
    Writes a record to the audit log resource table
    """
    try:
        AuditLogResource.create(db=db, data=audit_log_resource_data)
    except SQLAlchemyError as err:
        logger.debug(err)


async def get_client_user_id(db: Session, auth_token: str) -> str:
    """
    Attempts to retrieve a client user_id
    """
    stripped_token = auth_token.replace("Bearer ", "")
    _, client = extract_token_and_load_client(stripped_token, db)
    return client.user_id or "root"


async def extract_data_from_body(body: bytes) -> List:
    """
    Attempts to retrieve any fides_keys associated with
    the request found in the request body.
    """

    fides_keys: List[str] = []
    if body:
        body = json.loads(body)
        if isinstance(body, dict):
            fides_key = body.get("fides_key")
            if fides_key:
                fides_keys.append(fides_key)
        if isinstance(body, list):
            for body_item in body:
                if isinstance(body_item, dict):
                    fides_key = body_item.get("fides_key")
                    if fides_key:
                        fides_keys.append(fides_key)
    return fides_keys


async def set_body(request: Request, body: bytes) -> None:
    """
    Sets the body return type for use in middleware

    Required due to shortcomings in Starlette with awaiting request
    body in middleware

    Reference: https://github.com/tiangolo/fastapi/issues/394#issuecomment-883524819
    """

    async def receive() -> Message:
        return {"type": "http.request", "body": body}

    request._receive = receive  # pylint: disable=W0212


async def get_body(request: Request) -> bytes:
    """
    Awaits and sets the request body for use in middleware


    Required due to shortcomings in Starlette with awaiting request
    body in middleware

    Reference: https://github.com/tiangolo/fastapi/issues/394#issuecomment-883524819
    """
    body = await request.body()
    await set_body(request, body)
    return body



from datetime import datetime
from enum import Enum
from typing import Annotated, Any, Optional

from pydantic import ConfigDict, Field

from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.manual_tasks.manual_task_status import StatusType


class ManualTaskExecutionTiming(str, Enum):
    """Enum for when a manual task should be executed in the privacy request DAG."""

    pre_execution = "pre_execution"  # Execute before the main DAG
    post_execution = "post_execution"  # Execute after the main DAG
    parallel = "parallel"  # Execute in parallel with the main DAG


class ManualTaskType(str, Enum):
    """Enum for manual task types."""

    privacy_request = "privacy_request"
    # Add more task types as needed


class ManualTaskParentEntityType(str, Enum):
    """Enum for manual task parent entity types."""

    connection_config = (
        "connection_config"  # used for access and erasure privacy requests
    )
    # Add more parent entity types as needed


class ManualTaskEntityType(str, Enum):
    """Enum for manual task entity types."""

    privacy_request = "privacy_request"
    # Add more entity types as needed


class ManualTaskReferenceType(str, Enum):
    """Enum for manual task reference types."""

    privacy_request = "privacy_request"
    connection_config = "connection_config"
    manual_task_config = "manual_task_config"
    assigned_user = "assigned_user"  # Reference to the user assigned to the task
    # Add more reference types as needed


class ManualTaskLogStatus(str, Enum):
    """Enum for manual task log status."""

    created = "created"
    updated = "updated"
    in_progress = "in_progress"
    complete = "complete"
    error = "error"
    retrying = "retrying"
    paused = "paused"
    awaiting_input = "awaiting_input"


class ManualTaskResponse(FidesSchema):
    """Schema for manual task response."""

    model_config = ConfigDict(extra="forbid")

    id: Annotated[str, Field(..., description="Task ID")]
    parent_entity_id: Annotated[str, Field(..., description="Parent entity ID")]
    parent_entity_type: Annotated[
        ManualTaskParentEntityType, Field(..., description="Parent entity type")
    ]
    status: Annotated[StatusType, Field(..., description="Task status")]
    created_at: Annotated[datetime, Field(..., description="Creation timestamp")]
    updated_at: Annotated[datetime, Field(..., description="Last update timestamp")]


class ManualTaskCreate(FidesSchema):
    """Schema for creating a manual task."""

    model_config = ConfigDict(extra="forbid")

    parent_entity_id: Annotated[str, Field(..., description="Parent entity ID")]
    parent_entity_type: Annotated[
        ManualTaskParentEntityType, Field(..., description="Parent entity type")
    ]


class ManualTaskLogCreate(FidesSchema):
    """Schema for creating a manual task log entry."""

    model_config = ConfigDict(extra="forbid")

    task_id: Annotated[str, Field(..., description="ID of the task")]
    status: Annotated[ManualTaskLogStatus, Field(..., description="Log status")]
    message: Annotated[Optional[str], Field(None, description="Log message")]
    details: Annotated[
        Optional[dict[str, Any]], Field(None, description="Additional details")
    ]
    config_id: Annotated[Optional[str], Field(None, description="Configuration ID")]
    instance_id: Annotated[Optional[str], Field(None, description="Instance ID")]


class ManualTaskLogResponse(FidesSchema):
    """Schema for manual task log response."""

    model_config = ConfigDict(extra="forbid")

    id: Annotated[str, Field(..., description="Log ID")]
    task_id: Annotated[str, Field(..., description="Task ID")]
    status: Annotated[ManualTaskLogStatus, Field(..., description="Log status")]
    message: Annotated[Optional[str], Field(None, description="Log message")]
    details: Annotated[
        Optional[dict[str, Any]], Field(None, description="Additional details")
    ]
    config_id: Annotated[Optional[str], Field(None, description="Configuration ID")]
    instance_id: Annotated[Optional[str], Field(None, description="Instance ID")]
    created_at: Annotated[datetime, Field(..., description="Creation timestamp")]
    updated_at: Annotated[datetime, Field(..., description="Last update timestamp")]



"""Update table for twilio

Revision ID: 179f2bb623ae
Revises: 8f1a19465239
Create Date: 2022-10-31 18:19:26.845723

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "179f2bb623ae"
down_revision = "8f1a19465239"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "messagingconfig",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column(
            "service_type",
            sa.Enum(
                "MAILGUN", "TWILIO_TEXT", "TWILIO_EMAIL", name="messagingservicetype"
            ),
            nullable=False,
        ),
        sa.Column("details", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column(
            "secrets",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_messagingconfig_id"), "messagingconfig", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_messagingconfig_key"), "messagingconfig", ["key"], unique=True
    )
    op.create_index(
        op.f("ix_messagingconfig_name"), "messagingconfig", ["name"], unique=True
    )
    op.create_index(
        op.f("ix_messagingconfig_service_type"),
        "messagingconfig",
        ["service_type"],
        unique=True,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "emailconfig",
        sa.Column("id", sa.VARCHAR(length=255), autoincrement=False, nullable=False),
        sa.Column(
            "created_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("key", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.Column("name", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "service_type",
            postgresql.ENUM("MAILGUN", name="emailservicetype"),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column(
            "details",
            postgresql.JSONB(astext_type=sa.Text()),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column("secrets", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint("id", name="emailconfig_pkey"),
    )
    op.create_index(
        "ix_emailconfig_service_type", "emailconfig", ["service_type"], unique=False
    )
    op.create_index("ix_emailconfig_name", "emailconfig", ["name"], unique=False)
    op.create_index("ix_emailconfig_key", "emailconfig", ["key"], unique=False)
    op.create_index("ix_emailconfig_id", "emailconfig", ["id"], unique=False)
    op.drop_index(op.f("ix_messagingconfig_service_type"), table_name="messagingconfig")
    op.drop_index(op.f("ix_messagingconfig_name"), table_name="messagingconfig")
    op.drop_index(op.f("ix_messagingconfig_key"), table_name="messagingconfig")
    op.drop_index(op.f("ix_messagingconfig_id"), table_name="messagingconfig")
    op.drop_table("messagingconfig")
    # ### end Alembic commands ###



from os import getenv

DEFAULT_CONFIG_PATH = ".fides/fides.toml"
DEFAULT_CONFIG_PATH_ENV_VAR = "FIDES__CONFIG_PATH"


def replace_config_value(
    fides_directory_location: str, key: str, old_value: str, new_value: str
) -> None:
    """Use string replacment to update a value in the fides.toml"""

    # This matches the logic used in `docs.py`
    fides_dir_name = ".fides"
    fides_dir_path = f"{fides_directory_location}/{fides_dir_name}"
    config_file_name = "fides.toml"
    config_path = f"{fides_dir_path}/{config_file_name}"

    with open(config_path, "r", encoding="utf8") as config_file:
        previous_config = config_file.read()
        new_config = previous_config.replace(
            f"{key} = {old_value}", f"{key} = {new_value}"
        )

    with open(config_path, "w", encoding="utf8") as config_file:
        config_file.write(new_config)

    print(f"Config key: {key} value changed: {old_value} -> {new_value}")


def get_test_mode() -> bool:
    test_mode = getenv("FIDES__TEST_MODE", "").lower() == "true"
    return test_mode


def get_dev_mode() -> bool:
    dev_mode = getenv("FIDES__DEV_MODE", "").lower() == "true"
    return dev_mode


# NOTE: allowlist additions should be made with care!
# Any updates to this list _need_ to be reviewed by the Ethyca security team
CONFIG_KEY_ALLOWLIST = {
    "user": ["analytics_opt_out"],
    "logging": ["level"],
    "notifications": [
        "send_request_completion_notification",
        "send_request_receipt_notification",
        "send_request_review_notification",
        "notification_service_type",
        "enable_property_specific_messaging",
    ],
    "security": [
        "cors_origins",
        "cors_origin_regex",
        "encoding",
        "oauth_access_token_expire_minutes",
    ],
    "execution": [
        "task_retry_count",
        "task_retry_delay",
        "task_retry_backoff",
        "require_manual_request_approval",
        "subject_identity_verification_required",
    ],
    "storage": [
        "active_default_storage_type",
    ],
    "consent": ["override_vendor_purposes"],
    "admin_ui": ["enabled", "url", "error_notification_mode"],
}



"""remove user scopes

Revision ID: a0f219697fa0
Revises: 54102bba36de
Create Date: 2023-03-15 22:25:09.654149

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import text
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "a0f219697fa0"
down_revision = "54102bba36de"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_column("fidesuserpermissions", "scopes")
    """One last time - remove scopes from user clients"""
    bind = op.get_bind()
    bind.execute(
        text(
            """
                UPDATE client 
                SET scopes = '{}'
                FROM fidesuserpermissions
                WHERE fidesuserpermissions.user_id = client.user_id;
            """
        )
    )


def downgrade():
    op.add_column(
        "fidesuserpermissions",
        sa.Column(
            "scopes",
            postgresql.ARRAY(sa.VARCHAR()),
            server_default=sa.text("'{}'::character varying[]"),
            autoincrement=False,
            nullable=True,
        ),
    )



from typing import Any, Optional

from loguru import logger
from sqlalchemy.orm import Query, Session, selectinload

from fides.api.models.manual_tasks.manual_task import ManualTask
from fides.api.models.manual_tasks.manual_task_config import (
    ManualTaskConfig,
    ManualTaskConfigField,
)
from fides.api.schemas.manual_tasks.manual_task_config import (
    ManualTaskConfigResponse,
    ManualTaskConfigurationType,
)
from fides.service.manual_tasks.utils import validate_fields, with_task_logging


class ManualTaskConfigError(Exception):
    """Exception raised when a manual task config error occurs."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class ManualTaskConfigService:
    def __init__(self, db: Session):
        self.db = db

    def _create_log_data(
        self, task_id: str, config_id: Optional[str], details: dict[str, Any]
    ) -> dict[str, Any]:
        """Create standard log data structure."""
        return {"task_id": task_id, "config_id": config_id, "details": details}

    def _get_base_config_query(self) -> Query:
        """Get base config query with field definitions loaded."""
        return self.db.query(ManualTaskConfig).options(
            selectinload(ManualTaskConfig.field_definitions)
        )

    def _create_config_version(
        self,
        task: ManualTask,
        config_type: str,
        version: int,
        is_current: bool = True,
    ) -> ManualTaskConfig:
        """Create a new config version. This is a helper method that creates a new config
        version and sets it as the current version.

        Args:
            task: The task to create the config for
            config_type: The type of config to create
            version: The version number of the config
            is_current: Whether the config is the current version
        Returns:
            The new config
        """
        try:
            ManualTaskConfigurationType(config_type)
        except ManualTaskConfigError:
            raise ManualTaskConfigError(f"Invalid config type: {config_type}")

        # Set all existing versions to non-current
        if is_current:
            self.db.query(ManualTaskConfig).filter(
                ManualTaskConfig.task_id == task.id,
                ManualTaskConfig.config_type == config_type,
            ).update({"is_current": False})
            self.db.flush()

        return ManualTaskConfig.create(
            db=self.db,
            data={
                "task_id": task.id,
                "config_type": config_type,
                "version": version,
                "is_current": is_current,
            },
        )

    def _handle_field_updates(
        self,
        config: ManualTaskConfig,
        field_updates: Optional[list[dict[str, Any]]] = None,
        fields_to_remove: Optional[list[str]] = None,
        previous_config: Optional[ManualTaskConfig] = None,
    ) -> set[str]:
        """Handle field updates, removals, and recreation from previous version. This is a helper
        function that:
        - Creates new fields from field_updates
        - Recreates unmodified fields from previous version so they will remain in the config
        - Removes fields that are no longer in the config

        Args:
            config: The config to update
            field_updates: The fields to update
            fields_to_remove: The fields to remove
            previous_config: The previous config
        Returns:
            The modified keys
        """
        modified_keys = set(fields_to_remove or [])

        if field_updates:
            try:
                validate_fields(field_updates, is_submission=False)
            except ValueError as e:
                raise ManualTaskConfigError(f"Invalid field updates: {e}") from e

            fields_to_create = [
                {
                    "task_id": config.task_id,
                    "config_id": config.id,
                    "field_key": f["field_key"],
                    "field_type": f.get("field_type"),
                    "field_metadata": f.get("field_metadata", {}),
                }
                for f in field_updates
            ]
            if fields_to_create:
                self.db.bulk_insert_mappings(ManualTaskConfigField, fields_to_create)
                modified_keys.update(f["field_key"] for f in field_updates)

        if previous_config:
            # Recreate unmodified fields from previous version
            unmodified = [
                {
                    "task_id": config.task_id,
                    "config_id": config.id,
                    "field_key": f.field_key,
                    "field_type": f.field_type,
                    "field_metadata": f.field_metadata,
                }
                for f in previous_config.field_definitions
                if f.field_key not in modified_keys
            ]
            if unmodified:
                self.db.bulk_insert_mappings(ManualTaskConfigField, unmodified)

        self.db.flush()
        return modified_keys

    def to_response(self, config: ManualTaskConfig) -> ManualTaskConfigResponse:
        """Convert config model to response object."""
        return ManualTaskConfigResponse.model_validate(
            {
                "id": config.id,
                "task_id": config.task_id,
                "config_type": config.config_type,
                "version": config.version,
                "is_current": config.is_current,
                "fields": [
                    {
                        "field_key": f.field_key,
                        "field_type": f.field_type,
                        "field_metadata": f.field_metadata,
                    }
                    for f in config.field_definitions
                ],
                "created_at": config.created_at,
                "updated_at": config.updated_at,
            }
        )

    def get_current_config(
        self, task: ManualTask, config_type: str
    ) -> ManualTaskConfig:
        """Get current config for task by type."""
        config = (
            self._get_base_config_query()
            .filter(
                ManualTaskConfig.task_id == task.id,
                ManualTaskConfig.config_type == config_type,
                ManualTaskConfig.is_current.is_(True),
            )
            .first()
        )

        if not config:
            raise ManualTaskConfigError(
                f"No current config found for task {task.id} and type {config_type}"
            )
        return config

    def list_config_type_versions(
        self, task: ManualTask, config_type: str
    ) -> list[ManualTaskConfig]:
        """List all versions of a config type for a task."""
        return (
            self._get_base_config_query()
            .filter(
                ManualTaskConfig.task_id == task.id,
                ManualTaskConfig.config_type == config_type,
            )
            .order_by(ManualTaskConfig.version.desc())
            .all()
        )

    def get_config(
        self,
        task: ManualTask,
        config_type: str,
        field_id: str,
        config_id: str,
        field_key: str,
        version: int,
    ) -> Optional[ManualTaskConfig]:
        """Get config by various filters. This is a flexible lookup method that can find configs based on various filters.

        Args:
            task: The task to get the config for
            config_type: The type of config to get
            field_id: The ID of a field in the config
            config_id: The ID of the config
            field_key: The key of a field in the config
            version: The version number of the config
        Returns:
            The matching config if found, None otherwise
        """
        if not any([task, config_id, field_id, config_type]):
            logger.debug("No filters provided to get_config")
            return None

        query = self._get_base_config_query()
        if field_id or field_key:
            query = query.join(ManualTaskConfig.field_definitions)

        filters = [
            f
            for f in [
                task and ManualTaskConfig.task_id == task.id,
                config_id and ManualTaskConfig.id == config_id,
                field_id and ManualTaskConfigField.id == field_id,
                field_key and ManualTaskConfigField.field_key == field_key,
                version and ManualTaskConfig.version == version,
                config_type and ManualTaskConfig.config_type == config_type,
            ]
            if f
        ]

        return query.filter(*filters).first()

    @with_task_logging("Creating new configuration version")
    def create_new_version(
        self,
        task: ManualTask,
        config_type: str,
        field_updates: Optional[list[dict[str, Any]]] = None,
        fields_to_remove: Optional[list[str]] = None,
        previous_config: Optional[ManualTaskConfig] = None,
    ) -> tuple[ManualTaskConfig, dict[str, Any]]:
        """Create new version of configuration.

        Args:
            task: The task to create the config for
            config_type: The type of config to create
            field_updates: The fields to update
            fields_to_remove: The fields to remove
            previous_config: The previous config

        Returns:
            Tuple containing the new config and log data, the log data is
            captured by the with_task_logging decorator. and the new config is
            returned to the caller.
        """
        new_config = self._create_config_version(
            task, config_type, (previous_config.version + 1 if previous_config else 1)
        )

        self._handle_field_updates(
            new_config, field_updates, fields_to_remove, previous_config
        )

        return new_config, self._create_log_data(
            task.id,
            new_config.id,
            {
                "config_type": config_type,
                "version": new_config.version,
                "added_field_keys": [f.get("field_key") for f in (field_updates or [])],
                "removed_field_keys": fields_to_remove or [],
            },
        )

    @with_task_logging("Adding fields to configuration")
    def add_fields(
        self, task: ManualTask, config_type: str, fields: list[dict[str, Any]]
    ) -> tuple[ManualTaskConfig, dict[str, Any]]:
        """Add fields to configuration.

        Args:
            task: The task to add the fields to
            config_type: The type of config to add the fields to
            fields: The fields to add

        Returns:
            Tuple containing the new config and log data, the log data is
            captured by the with_task_logging decorator. and the new config is
            returned to the caller.
        """
        current = self.get_current_config(task, config_type)
        self.create_new_version(task, config_type, fields, previous_config=current)
        new_config = self.get_current_config(task, config_type)

        return new_config, self._create_log_data(
            task.id,
            current.id,
            {
                "config_type": config_type,
                "previous_version": current.version,
                "added_field_keys": [f.get("field_key") for f in fields],
                "new_config_id": new_config.id,
                "new_config_version": new_config.version,
            },
        )

    @with_task_logging("Removing fields from configuration")
    def remove_fields(
        self, task: ManualTask, config_type: str, field_keys: list[str]
    ) -> tuple[ManualTaskConfig, dict[str, Any]]:
        """Remove fields from configuration.

        Args:
            task: The task to remove the fields from
            config_type: The type of config to remove the fields from
            field_keys: The keys of the fields to remove

        Returns:
            Tuple containing the new config and log data, the log data is
            captured by the with_task_logging decorator. and the new config is
            returned to the caller.
        """
        current = self.get_current_config(task, config_type)
        self.create_new_version(
            task, config_type, fields_to_remove=field_keys, previous_config=current
        )
        new_config = self.get_current_config(task, config_type)

        return new_config, self._create_log_data(
            task.id,
            current.id,
            {
                "config_type": config_type,
                "version": current.version,
                "deleted_field_keys": field_keys,
                "new_config_id": new_config.id,
                "new_config_version": new_config.version,
            },
        )

    @with_task_logging("Deleting Manual Task configuration")
    def delete_config(
        self, task: ManualTask, config_id: str
    ) -> tuple[ManualTaskConfig, dict[str, Any]]:
        """Delete config for task.

        Args:
            task: The task to delete the config for
            config_id: The ID of the config to delete

        Returns:
            Tuple containing the deleted config and log data, the log data is
            captured by the with_task_logging decorator. and the deleted config is
            returned to the caller.
        """
        config = self.db.query(ManualTaskConfig).filter_by(id=config_id).first()
        if not config:
            raise ManualTaskConfigError(f"Config with ID {config_id} not found")

        log_data = self._create_log_data(
            task.id,
            None,
            {
                "config_type": config.config_type,
                "version": config.version,
                "deleted_config_id": config_id,
            },
        )
        config.delete(self.db)
        return config, log_data



from __future__ import annotations

from io import BytesIO
from typing import IO, Any, Dict, Optional, Tuple, Union

from boto3.s3.transfer import TransferConfig
from botocore.exceptions import ClientError, ParamValidationError
from fideslang.validation import AnyHttpUrlString
from loguru import logger

from fides.api.schemas.storage.storage import StorageSecrets
from fides.api.service.storage.util import (
    LARGE_FILE_THRESHOLD,
    get_allowed_file_type_or_raise,
)
from fides.api.util.aws_util import get_s3_client
from fides.config import CONFIG


def maybe_get_s3_client(
    auth_method: str, storage_secrets: Dict[StorageSecrets, Any]
) -> Any:
    """
    Returns an S3 client if the client can be created successfully, otherwise raises an exception.
    """
    try:
        return get_s3_client(auth_method, storage_secrets)
    except ClientError as e:
        logger.error(
            "Encountered error while uploading and generating link for S3 object: {}", e
        )
        raise e
    except ParamValidationError as e:
        raise ValueError(f"The parameters you provided are incorrect: {e}")


def create_presigned_url_for_s3(
    s3_client: Any, bucket_name: str, file_key: str, ttl_seconds: Optional[int] = None
) -> AnyHttpUrlString:
    """
    Generates a presigned URL to share an S3 object

    :param s3_client: s3 base client
    :param bucket_name: string
    :param file_key: string
    :return: Presigned URL as string.
    """
    params = {"Bucket": bucket_name, "Key": file_key}
    if ttl_seconds:
        if ttl_seconds > 604800:
            raise ValueError("TTL must be less than 7 days")
        expires_in = ttl_seconds
    else:
        expires_in = CONFIG.security.subject_request_download_link_ttl_seconds
    response = s3_client.generate_presigned_url(
        "get_object",
        Params=params,
        ExpiresIn=expires_in,
    )

    # The response contains the presigned URL
    return response


def get_file_size(s3_client: Any, bucket_name: str, file_key: str) -> int:
    """
    Returns the size of a file in bytes.
    """
    return s3_client.head_object(Bucket=bucket_name, Key=file_key)["ContentLength"]


def generic_upload_to_s3(  # pylint: disable=R0913
    storage_secrets: Dict[StorageSecrets, Any],
    bucket_name: str,
    file_key: str,
    auth_method: str,
    document: IO[bytes],
    size_threshold: int = LARGE_FILE_THRESHOLD,  # 25 MB threshold
) -> Tuple[int, AnyHttpUrlString]:
    """
    Uploads file like objects to S3.
    Handles both small and large uploads.

    :param storage_secrets: S3 storage secrets
    :param bucket_name: Name of the S3 bucket
    :param file_key: Key of the file in the bucket
    :param auth_method: Authentication method for S3
    :param document: File contents to upload
    """
    logger.info("Starting S3 Upload of {}", file_key)

    # Validate that the document is a file-like object
    if not hasattr(document, "read") or not hasattr(document, "seek"):
        raise TypeError(
            f"The 'document' parameter must be a file-like object supporting 'read' and 'seek'. "
            f"Received: {type(document)}, {document}"
        )

    # Ensure the file pointer is at the beginning
    try:
        document.seek(0)
    except Exception as e:
        raise ValueError(f"Failed to reset file pointer for document: {e}")

    s3_client = maybe_get_s3_client(auth_method, storage_secrets)

    # Define a transfer configuration for multipart uploads
    transfer_config = TransferConfig(
        multipart_threshold=size_threshold,
        multipart_chunksize=size_threshold,
    )

    # Use upload_fileobj for efficient uploads (handles both small and large files)
    try:
        file_type: str = get_allowed_file_type_or_raise(file_key)
        # Use upload_fileobj for efficient uploads (handles both small and large files)
        s3_client.upload_fileobj(
            Fileobj=document,
            Bucket=bucket_name,
            Key=file_key,
            Config=transfer_config,
            ExtraArgs={"ContentType": file_type} if file_type else {},
        )
        file_size = get_file_size(s3_client, bucket_name, file_key)
    except Exception as e:
        logger.error(f"Failed to upload file {file_key} to bucket {bucket_name}: {e}")
        raise e  # Re-raise the exception if you want it to propagate

    # Generate a presigned URL for the uploaded file
    logger.info(f"Successfully uploaded file {file_key} to bucket {bucket_name}")
    return file_size, create_presigned_url_for_s3(s3_client, bucket_name, file_key)


def generic_retrieve_from_s3(
    storage_secrets: Dict[StorageSecrets, Any],
    bucket_name: str,
    file_key: str,
    auth_method: str,
    get_content: bool = False,
    ttl_seconds: Optional[int] = None,
) -> Tuple[int, Union[str, IO[bytes]]]:
    """
    Retrieves a file from S3 and returns its size and either a presigned URL or the actual content.

    Args:
        storage_secrets: Dictionary containing S3 credentials
        bucket_name: Name of the S3 bucket
        file_key: Key of the file in the bucket
        auth_method: Authentication method to use
        get_content: If True, returns the actual content instead of a presigned URL

    Returns:
        Tuple containing (file_size, presigned_url_or_content)
    """
    logger.info("Retrieving S3 object: {}", file_key)
    s3_client = get_s3_client(auth_method, storage_secrets)

    try:
        # Get file size using head_object
        size_response = s3_client.head_object(Bucket=bucket_name, Key=file_key)
        # If the file is less than 25MB, we can get the content otherwise return the presigned URL
        if get_content and size_response["ContentLength"] <= LARGE_FILE_THRESHOLD:
            # Get the actual content using download_fileobj
            file_obj = BytesIO()
            s3_client.download_fileobj(
                Bucket=bucket_name, Key=file_key, Fileobj=file_obj
            )
            file_obj.seek(0)  # Reset file pointer to beginning
            return int(size_response["ContentLength"]), file_obj

        # Get presigned URL
        presigned_url = create_presigned_url_for_s3(
            s3_client, bucket_name, file_key, ttl_seconds
        )
        return int(size_response["ContentLength"]), str(presigned_url)
    except ClientError as e:
        logger.error(f"Error retrieving file from S3: {e}")
        raise e


def generic_delete_from_s3(
    storage_secrets: Dict[StorageSecrets, Any],
    bucket_name: str,
    file_key: str,
    auth_method: str,
) -> None:
    """
    Deletes arbitrary data from s3. If file_key ends with a filename, deletes just that file.
    If file_key ends with a '/', deletes all objects with that prefix.

    :param storage_secrets: S3 storage secrets
    :param bucket_name: Name of the S3 bucket
    :param file_key: Key of the file in the bucket
    :param auth_method: Authentication method for S3
    """
    logger.info("Starting S3 Delete of {}", file_key)

    s3_client = maybe_get_s3_client(auth_method, storage_secrets)
    try:
        # If the file_key ends with a '/', it's a folder prefix
        if file_key.endswith("/"):
            # List all objects with the prefix, handling pagination
            paginator = s3_client.get_paginator("list_objects_v2")
            for page in paginator.paginate(Bucket=bucket_name, Prefix=file_key):
                if "Contents" in page:
                    for obj in page["Contents"]:
                        s3_client.delete_object(Bucket=bucket_name, Key=obj["Key"])
            return

        # Delete single object
        s3_client.delete_object(Bucket=bucket_name, Key=file_key)
    except Exception as e:
        logger.error("Encountered error while deleting s3 object: {}", e)
        raise e



"""update privacy request relationships

Revision ID: 1088e8353890
Revises: d9237a0c0d5a
Create Date: 2024-12-26 22:38:37.905571

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "1088e8353890"
down_revision = "d9237a0c0d5a"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint(
        "custom_privacy_request_field_privacy_request_id_fkey",
        "custom_privacy_request_field",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "custom_privacy_request_field_privacy_request_id_fkey",
        "custom_privacy_request_field",
        "privacyrequest",
        ["privacy_request_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.drop_constraint(
        "providedidentity_privacy_request_id_fkey",
        "providedidentity",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "providedidentity_privacy_request_id_fkey",
        "providedidentity",
        "privacyrequest",
        ["privacy_request_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.drop_constraint(
        "privacyrequesterror_privacy_request_id_fkey",
        "privacyrequesterror",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "privacyrequesterror_privacy_request_id_fkey",
        "privacyrequesterror",
        "privacyrequest",
        ["privacy_request_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )


def downgrade():
    op.drop_constraint(
        "providedidentity_privacy_request_id_fkey",
        "providedidentity",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "providedidentity_privacy_request_id_fkey",
        "providedidentity",
        "privacyrequest",
        ["privacy_request_id"],
        ["id"],
    )
    op.drop_constraint(
        "custom_privacy_request_field_privacy_request_id_fkey",
        "custom_privacy_request_field",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "custom_privacy_request_field_privacy_request_id_fkey",
        "custom_privacy_request_field",
        "privacyrequest",
        ["privacy_request_id"],
        ["id"],
    )
    op.drop_constraint(
        "privacyrequesterror_privacy_request_id_fkey",
        "privacyrequesterror",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "privacyrequesterror_privacy_request_id_fkey",
        "privacyrequesterror",
        "privacyrequest",
        ["privacy_request_id"],
        ["id"],
    )



"""add parent ID to privacy notice

Revision ID: 2394e4f69a49
Revises: 9de4bb76307a
Create Date: 2024-09-09 21:28:20.286331

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2394e4f69a49"
down_revision = "9de4bb76307a"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("privacynotice", sa.Column("parent_id", sa.String(), nullable=True))
    op.create_foreign_key(
        None,
        "privacynotice",
        "privacynotice",
        ["parent_id"],
        ["id"],
        ondelete="SET NULL",
    )


def downgrade():
    op.drop_constraint(None, "privacynotice", type_="foreignkey")
    op.drop_column("privacynotice", "parent_id")



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from loguru import logger
from redis.lock import Lock

from fides.api.util.cache import FidesopsRedis, get_cache


def get_redis_lock(lock_key: str, timeout: int) -> Lock:
    """
    Creates and returns a Redis lock object.
    """
    redis: FidesopsRedis = get_cache()
    return redis.lock(lock_key, timeout=timeout)


@contextmanager
def redis_lock(lock_key: str, timeout: int) -> Generator[Lock | None, None, None]:
    """
    A context manager for acquiring a Redis lock.
    If the lock is acquired, it yields the lock object.
    If the lock is not acquired, it yields None.
    The lock is automatically released on exiting the context.
    """
    lock = get_redis_lock(lock_key, timeout)

    if not lock.acquire(blocking=False):
        logger.info(
            f"Another process is already running for lock '{lock_key}'. Skipping this execution."
        )
        yield None
        return

    try:
        logger.info(f"Acquired lock for '{lock_key}'.")
        yield lock
    finally:
        # Always release the lock if we own it
        if lock.owned():
            lock.release()
            logger.info(f"Released lock for '{lock_key}'.")



# pylint: disable=useless-import-alias

# Because these modules are imported into the __init__.py and used elsewhere they need
# to be explicitly exported in order to prevent implicit reexport errors in mypy.
# This is done by importing "as": `from fides.module import MyClass as MyClass`.

from typing import Any, Dict, Optional, Union

from fides.api.models.connectionconfig import ConnectionType
from fides.api.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema as ConnectionConfigSecretsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_attentive import (
    AttentiveSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_bigquery import (
    BigQueryDocsSchema as BigQueryDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_bigquery import (
    BigQuerySchema as BigQuerySchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_datahub import (
    DatahubDocsSchema as DatahubDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_datahub import (
    DatahubSchema as DatahubSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_dynamic_erasure_email import (
    DynamicErasureEmailDocsSchema as DynamicErasureEmailDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_dynamic_erasure_email import (
    DynamicErasureEmailSchema as DynamicErasureEmailSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_dynamodb import (
    DynamoDBDocsSchema as DynamoDBDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_dynamodb import (
    DynamoDBSchema as DynamoDBSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_email import (
    EmailDocsSchema as EmailDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_email import (
    EmailSchema as EmailSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_email import (
    ExtendedEmailSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_fides import (
    FidesConnectorSchema,
    FidesDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_google_cloud_sql_mysql import (
    GoogleCloudSQLMySQLDocsSchema as GoogleCloudSQLMySQLDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_google_cloud_sql_mysql import (
    GoogleCloudSQLMySQLSchema as GoogleCloudSQLMySQLSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_google_cloud_sql_postgres import (
    GoogleCloudSQLPostgresDocsSchema as GoogleCloudSQLPostgresDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_google_cloud_sql_postgres import (
    GoogleCloudSQLPostgresSchema as GoogleCloudSQLPostgresSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_manual_webhook import (
    ManualWebhookDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_manual_webhook import (
    ManualWebhookSchema as ManualWebhookSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_mariadb import (
    MariaDBDocsSchema as MariaDBDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_mariadb import (
    MariaDBSchema as MariaDBSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_mongodb import (
    MongoDBDocsSchema as MongoDBDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_mongodb import (
    MongoDBSchema as MongoDBSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_mssql import (
    MicrosoftSQLServerSchema as MicrosoftSQLServerSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_mssql import (
    MSSQLDocsSchema as MSSQLDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_mysql import (
    MySQLDocsSchema as MySQLDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_mysql import (
    MySQLSchema as MySQLSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_okta import (
    OktaDocsSchema as OktaDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_okta import (
    OktaSchema as OktaSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_postgres import (
    PostgreSQLDocsSchema as PostgreSQLDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_postgres import (
    PostgreSQLSchema as PostgreSQLSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_rds_mysql import (
    RDSMySQLDocsSchema as RDSMySQLDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_rds_mysql import (
    RDSMySQLSchema as RDSMySQLSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_rds_postgres import (
    RDSPostgresDocsSchema as RDSPostgresDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_rds_postgres import (
    RDSPostgresSchema as RDSPostgresSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_redshift import (
    RedshiftDocsSchema as RedshiftDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_redshift import (
    RedshiftSchema as RedshiftSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_s3 import (
    S3DocsSchema as S3DocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_s3 import (
    S3Schema as S3Schema,
)
from fides.api.schemas.connection_configuration.connection_secrets_saas import (
    SaaSSchema as SaaSSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_saas import (
    SaaSSchemaFactory as SaaSSchemaFactory,
)
from fides.api.schemas.connection_configuration.connection_secrets_scylla import (
    ScyllaDocsSchema,
    ScyllaSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_snowflake import (
    SnowflakeDocsSchema as SnowflakeDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_snowflake import (
    SnowflakeSchema as SnowflakeSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_sovrn import (
    SovrnDocsSchema,
    SovrnSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_timescale import (
    TimescaleDocsSchema as TimescaleDocsSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_timescale import (
    TimescaleSchema as TimescaleSchema,
)
from fides.api.schemas.connection_configuration.connection_secrets_website import (
    WebsiteSchema,
)
from fides.api.schemas.connection_configuration.connections_secrets_https import (
    HttpsSchema as HttpsSchema,
)
from fides.api.schemas.saas.saas_config import SaaSConfig as SaaSConfig

secrets_schemas: Dict[str, Any] = {
    ConnectionType.attentive_email.value: AttentiveSchema,
    ConnectionType.bigquery.value: BigQuerySchema,
    ConnectionType.datahub.value: DatahubSchema,
    ConnectionType.dynamic_erasure_email.value: DynamicErasureEmailSchema,
    ConnectionType.dynamodb.value: DynamoDBSchema,
    ConnectionType.fides.value: FidesConnectorSchema,
    ConnectionType.generic_consent_email.value: ExtendedEmailSchema,
    ConnectionType.generic_erasure_email.value: EmailSchema,
    ConnectionType.google_cloud_sql_mysql.value: GoogleCloudSQLMySQLSchema,
    ConnectionType.google_cloud_sql_postgres.value: GoogleCloudSQLPostgresSchema,
    ConnectionType.https.value: HttpsSchema,
    ConnectionType.manual_webhook.value: ManualWebhookSchema,
    ConnectionType.mariadb.value: MariaDBSchema,
    ConnectionType.mongodb.value: MongoDBSchema,
    ConnectionType.mssql.value: MicrosoftSQLServerSchema,
    ConnectionType.okta.value: OktaSchema,
    ConnectionType.mysql.value: MySQLSchema,
    ConnectionType.postgres.value: PostgreSQLSchema,
    ConnectionType.rds_mysql.value: RDSMySQLSchema,
    ConnectionType.rds_postgres.value: RDSPostgresSchema,
    ConnectionType.redshift.value: RedshiftSchema,
    ConnectionType.s3.value: S3Schema,
    ConnectionType.saas.value: SaaSSchema,
    ConnectionType.scylla.value: ScyllaSchema,
    ConnectionType.snowflake.value: SnowflakeSchema,
    ConnectionType.sovrn.value: SovrnSchema,
    ConnectionType.timescale.value: TimescaleSchema,
    ConnectionType.website.value: WebsiteSchema,
}


def get_connection_secrets_schema(
    connection_type: str, saas_config: Optional[SaaSConfig]
) -> ConnectionConfigSecretsSchema:
    """
    Returns a validation schema for the connection "secrets" depending on the connection_type.

    For example, a "postgres" connection_type would need to have its secrets
    validated against a PostgreSQL schema.
    """
    if connection_type == "saas" and not saas_config:
        raise ValueError(
            "A SaaS config to validate the secrets is required for a saas connection"
        )
    try:
        schema = (
            SaaSSchemaFactory(saas_config).get_saas_schema()
            if saas_config
            else secrets_schemas[connection_type]
        )
        return schema
    except KeyError:
        raise NotImplementedError(
            f"Add {connection_type} to the 'secrets_schema' mapping."
        )


# These schemas are used to generate the documentation for the connection secrets
# but validation is intentionally postponed due to the NoValidationSchema.
# Creating the actual connection secrets schemas happens later once we know
# what type of schema we should validate against.
connection_secrets_schemas = Union[
    BigQueryDocsSchema,
    DatahubDocsSchema,
    DynamicErasureEmailDocsSchema,
    DynamoDBDocsSchema,
    EmailDocsSchema,
    FidesDocsSchema,
    GoogleCloudSQLMySQLDocsSchema,
    GoogleCloudSQLPostgresDocsSchema,
    ManualWebhookDocsSchema,
    MariaDBDocsSchema,
    MongoDBDocsSchema,
    MSSQLDocsSchema,
    MySQLDocsSchema,
    OktaDocsSchema,
    PostgreSQLDocsSchema,
    RDSMySQLDocsSchema,
    RDSPostgresDocsSchema,
    RedshiftDocsSchema,
    S3DocsSchema,
    SaaSSchema,
    ScyllaDocsSchema,
    SnowflakeDocsSchema,
    SovrnDocsSchema,
    TimescaleDocsSchema,
    WebsiteSchema,
]



import glob
import re
from hashlib import sha1
from os import getenv
from os.path import isfile
from pathlib import Path
from typing import Any, Dict, Iterator, List

import sqlalchemy
import toml
from fideslang.models import DatasetField, FidesModel
from loguru import logger
from pydantic import BaseModel, ValidationError
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError

from fides.common.utils import echo_red
from fides.connectors.models import ConnectorAuthFailureException

logger.bind(name="server_api")


class Credentials(BaseModel):
    """
    User credentials for the CLI.
    """

    username: str
    user_id: str
    access_token: str


def get_db_engine(connection_string: str) -> Engine:
    """
    Use SQLAlchemy to create a DB engine.
    """

    # Pymssql doesn't support this arg
    connect_args: Dict[str, Any] = (
        {"connect_timeout": 10} if "pymssql" not in connection_string else {}
    )
    if "redshift" in connection_string:
        connect_args["sslmode"] = "prefer"
        connect_args["connect_timeout"] = 60
    try:
        engine = sqlalchemy.create_engine(connection_string, connect_args=connect_args)
    except Exception as err:
        raise Exception("Failed to create engine!") from err

    try:
        with engine.begin() as connection:
            connection.execute("SELECT 1")
    except Exception as err:
        raise Exception(f"Database connection failed with engine:\n{engine}!") from err
    return engine


def validate_db_engine(connection_string: str) -> None:
    """
    Use SQLAlchemy to create a DB engine.
    """
    # Pymssql doesn't support this arg
    connect_args: Dict[str, Any] = (
        {"connect_timeout": 10} if "pymssql" not in connection_string else {}
    )
    if "redshift" in connection_string:
        connect_args["sslmode"] = "prefer"
        connect_args["connect_timeout"] = 60
    try:
        engine = sqlalchemy.create_engine(connection_string, connect_args=connect_args)
        with engine.begin() as connection:
            connection.execute("SELECT 1")
    except SQLAlchemyError as error:
        raise ConnectorAuthFailureException(error)


def get_all_level_fields(fields: list) -> Iterator[DatasetField]:
    """
    Traverses all levels of fields that exist in a dataset
    returning them for individual evaluation.
    """
    for field in fields:
        yield field
        if isinstance(field, dict):
            if field["fields"]:
                yield from get_all_level_fields(field["fields"])
        else:
            if field.fields:
                yield from get_all_level_fields(field.fields)


def get_manifest_list(manifests_dir: str) -> List[str]:
    """Get a list of manifest files from the manifest directory."""
    yml_endings = ["yml", "yaml"]

    if isfile(manifests_dir) and manifests_dir.split(".")[-1] in yml_endings:
        return [manifests_dir]

    manifest_list = []
    for yml_ending in yml_endings:
        manifest_list += glob.glob(f"{manifests_dir}/**/*.{yml_ending}", recursive=True)

    return manifest_list


def check_fides_key(proposed_fides_key: str) -> str:
    """
    A helper function to automatically sanitize
    an invalid FidesKey.
    """
    try:
        FidesModel(fides_key=proposed_fides_key)
        return proposed_fides_key
    except ValidationError as error:
        echo_red(error)
        return sanitize_fides_key(proposed_fides_key)


def sanitize_fides_key(proposed_fides_key: str) -> str:
    """
    Attempts to manually sanitize a fides key if restricted
    characters are discovered.
    """
    sanitized_fides_key = re.sub(r"[^a-zA-Z0-9_.-]", "_", proposed_fides_key)
    return sanitized_fides_key


def generate_unique_fides_key(
    proposed_fides_key: str, database_host: str, database_name: str
) -> str:
    """
    Uses host and name to generate a UUID to be
    appended to the fides_key of a dataset
    """
    fides_key_uuid = sha1(
        "-".join([database_host, database_name, proposed_fides_key]).encode()
    )
    return f"{proposed_fides_key}_{fides_key_uuid.hexdigest()[:10]}"


def git_is_dirty(dir_to_check: str = ".") -> bool:
    """
    Checks to see if the local repo has unstaged changes.
    Can also specify a directory to check.
    """

    try:
        from git.repo import Repo
        from git.repo.fun import is_git_dir
    except ImportError:
        print("Git executable not detected, skipping git check...")
        return False

    git_dir_path = ".git/"
    if not is_git_dir(git_dir_path):
        print(f"No git repo detected at '{git_dir_path}', skipping git check...")
        return False

    repo = Repo()
    git_session = repo.git()

    dirty_phrases = ["Changes not staged for commit:", "Untracked files:"]
    git_status = git_session.status(dir_to_check).split("\n")
    is_dirty = any(phrase in git_status for phrase in dirty_phrases)
    return is_dirty


def write_credentials_file(credentials: Credentials, credentials_path: str) -> str:
    """Write the user credentials file."""
    with open(credentials_path, "w", encoding="utf-8") as credentials_file:
        credentials_file.write(toml.dumps(credentials.model_dump(mode="json")))
    return credentials_path


def get_credentials_path() -> str:
    """
    Returns the default credentials path or the path set as an environment variable.
    """
    default_credentials_file_path = f"{str(Path.home())}/.fides_credentials"
    credentials_path = getenv("FIDES_CREDENTIALS_PATH", default_credentials_file_path)
    return credentials_path


def read_credentials_file(
    credentials_path: str,
) -> Credentials:
    """Read and return the credentials file."""
    if not isfile(credentials_path):
        raise FileNotFoundError
    with open(credentials_path, "r", encoding="utf-8") as credentials_file:
        credentials = Credentials.model_validate(toml.load(credentials_file))
    return credentials


def create_auth_header(access_token: str) -> Dict[str, str]:
    """Given an access token, create an auth header."""
    auth_header = {
        "Authorization": f"Bearer {access_token}",
    }
    return auth_header


def get_auth_header(verbose: bool = True) -> Dict[str, str]:
    """
    Executes all of the logic required to form a valid auth header.
    """
    credentials_path = get_credentials_path()
    try:
        credentials = read_credentials_file(credentials_path=credentials_path)
    except FileNotFoundError:
        if verbose:
            echo_red("No credentials file found.")
        raise SystemExit(1)

    access_token = credentials.access_token
    auth_header = create_auth_header(access_token)
    return auth_header



from typing import Literal, Set, Tuple

from fides.api.schemas.connection_configuration import secrets_schemas
from fides.api.schemas.namespace_meta.namespace_meta import NamespaceMeta


class SnowflakeNamespaceMeta(NamespaceMeta):
    """
    Represents the namespace structure for Snowflake queries.

    Attributes:
        database_name (str): Name of the specific Snowflake database.
        schema (str): The schema within the database.
    """

    connection_type: Literal["snowflake"] = "snowflake"
    database_name: str
    schema: str  # type: ignore[assignment]

    @classmethod
    def get_fallback_secret_fields(cls) -> Set[Tuple]:
        """
        The required connection config secrets when namespace metadata is missing.
        For Snowflake, database_name and schema_name are required.
        """
        schema = secrets_schemas["snowflake"]
        return {
            ("database_name", schema.model_fields["database_name"].title),
            ("schema_name", schema.model_fields["schema_name"].title),
        }



from datetime import datetime, timezone
from typing import TYPE_CHECKING, Optional

from sqlalchemy import Column, DateTime, ForeignKey, Index, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func

from fides.api.db.base_class import Base, FidesBase
from fides.api.db.util import EnumColumn
from fides.api.models.manual_tasks.manual_task_config import ManualTaskConfigField
from fides.api.schemas.manual_tasks.manual_task_schemas import ManualTaskEntityType
from fides.api.schemas.manual_tasks.manual_task_status import (
    StatusTransitionMixin,
    StatusType,
)

if TYPE_CHECKING:  # pragma: no cover
    from fides.api.models.attachment import Attachment  # pragma: no cover
    from fides.api.models.fides_user import FidesUser  # pragma: no cover
    from fides.api.models.manual_tasks.manual_task import ManualTask  # pragma: no cover
    from fides.api.models.manual_tasks.manual_task_config import (
        ManualTaskConfig,  # pragma: no cover
    )
    from fides.api.models.manual_tasks.manual_task_log import (
        ManualTaskLog,  # pragma: no cover; pragma: no cover
    )


class ManualTaskInstance(Base, StatusTransitionMixin):
    """Model for tracking task status per entity instance.

    This model implements StatusTransitionProtocol through the StatusTransitionMixin.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task_instance"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    # Database columns
    task_id: Column[str] = Column(
        String,
        ForeignKey(
            "manual_task.id",
            name="manual_task_instance_task_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    config_id: Column[str] = Column(
        String,
        ForeignKey(
            "manual_task_config.id",
            name="manual_task_instance_config_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    # entity id is the entity that the instance relates to
    # (e.g. a privacy request is an entity that has its own manual task instance)
    entity_id: Column[str] = Column(String, nullable=False)
    entity_type: Column[ManualTaskEntityType] = Column(
        EnumColumn(ManualTaskEntityType), nullable=False
    )
    # ingnore[assignment] because the mypy and sqlalchemy types mismatch
    # upgrading to 2.0 allows mapping which provides better type safety visibility.
    status: Column[StatusType] = Column(EnumColumn(StatusType), nullable=False, default=StatusType.pending)  # type: ignore[assignment]
    completed_at: Column[Optional[datetime]] = Column(DateTime, nullable=True)  # type: ignore[assignment]
    completed_by_id: Column[Optional[str]] = Column(String, nullable=True)  # type: ignore[assignment]
    due_date: Column[Optional[datetime]] = Column(DateTime, nullable=True)

    __table_args__ = (
        Index("ix_manual_task_instance_completed_at", "completed_at"),
        Index("ix_manual_task_instance_config_id", "config_id"),
        Index("ix_manual_task_instance_entity", "entity_type", "entity_id"),
        Index("ix_manual_task_instance_entity_id", "entity_id"),
        Index("ix_manual_task_instance_entity_type", "entity_type"),
        Index("ix_manual_task_instance_status", "status"),
        Index("ix_manual_task_instance_task_id", "task_id"),
    )

    # Relationships
    task = relationship("ManualTask", back_populates="instances")
    config = relationship("ManualTaskConfig", back_populates="instances")
    submissions = relationship(
        "ManualTaskSubmission",
        back_populates="instance",
        cascade="all, delete-orphan",
        uselist=True,
    )
    logs = relationship(
        "ManualTaskLog",
        back_populates="instance",
        primaryjoin="ManualTaskInstance.id == ManualTaskLog.instance_id",
        cascade="all, delete-orphan",
        order_by="ManualTaskLog.created_at",
        uselist=True,
    )
    attachments = relationship(
        "Attachment",
        secondary="attachment_reference",
        primaryjoin="and_(ManualTaskInstance.id == ManualTaskSubmission.instance_id, "
        "ManualTaskSubmission.id == AttachmentReference.reference_id, "
        "AttachmentReference.reference_type == 'manual_task_submission')",
        secondaryjoin="Attachment.id == AttachmentReference.attachment_id",
        order_by="Attachment.created_at",
        viewonly=True,
        uselist=True,
    )

    @property
    def required_fields(self) -> list["ManualTaskConfigField"]:
        """Get all required fields."""
        return [
            field
            for field in self.config.field_definitions
            if field.field_metadata.get("required", False)
        ]

    @property
    def incomplete_fields(self) -> list["ManualTaskConfigField"]:
        """Get all fields that haven't been completed yet.
        A field is considered incomplete if:
        1. It's required and has no submission
        Returns:
            list[ManualTaskConfigField]: List of incomplete fields
        """
        return [
            field
            for field in self.required_fields
            if not self.get_submission_for_field(field.id)
        ]

    @property
    def completed_fields(self) -> list["ManualTaskConfigField"]:
        """Get all fields that have been completed."""
        return [
            field
            for field in self.config.field_definitions
            if field.field_metadata.get("required", False)
            and self.get_submission_for_field(field.id)
        ]

    def get_submission_for_field(
        self, field_id: str
    ) -> Optional["ManualTaskSubmission"]:
        """Get the submission for a specific field.

        Args:
            field_id: The ID of the field to get the submission for

        Returns:
            Optional[ManualTaskSubmission]: The submission for the field, or None if no submission exists
        """
        return next(
            (
                submission
                for submission in self.submissions
                if submission.field_id == field_id
            ),
            None,
        )


class ManualTaskSubmission(Base):
    """Model for storing user submissions.
    Each submission represents data for a single field.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task_submission"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    # Database columns
    task_id = Column(
        String,
        ForeignKey(
            "manual_task.id",
            name="manual_task_submission_task_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    config_id = Column(
        String,
        ForeignKey(
            "manual_task_config.id",
            name="manual_task_submission_config_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    field_id = Column(
        String,
        ForeignKey(
            "manual_task_config_field.id",
            name="manual_task_submission_field_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    instance_id = Column(
        String,
        ForeignKey(
            "manual_task_instance.id",
            name="manual_task_submission_instance_id_fkey",
            ondelete="CASCADE",
        ),
        nullable=False,
    )
    submitted_by = Column(
        String,
        ForeignKey(
            "fidesuser.id",
            name="manual_task_submission_submitted_by_fkey",
            ondelete="SET NULL",
        ),
        nullable=True,
    )
    submitted_at = Column(DateTime, default=datetime.now(timezone.utc), nullable=False)
    data = Column(JSONB, nullable=False)

    __table_args__ = (
        Index("ix_manual_task_submission_config_id", "config_id"),
        Index("ix_manual_task_submission_field_id", "field_id"),
        Index("ix_manual_task_submission_instance_field", "instance_id", "field_id"),
        Index("ix_manual_task_submission_instance_id", "instance_id"),
        Index("ix_manual_task_submission_submitted_at", "submitted_at"),
        Index("ix_manual_task_submission_submitted_by", "submitted_by"),
        Index("ix_manual_task_submission_task_id", "task_id"),
    )

    # Relationships
    task = relationship("ManualTask", back_populates="submissions", viewonly=True)
    config = relationship(
        "ManualTaskConfig", back_populates="submissions", viewonly=True
    )
    field = relationship(
        "ManualTaskConfigField", back_populates="submissions", viewonly=True
    )
    instance = relationship(
        "ManualTaskInstance", back_populates="submissions", viewonly=True
    )
    attachments = relationship(
        "Attachment",
        secondary="attachment_reference",
        primaryjoin="and_(ManualTaskSubmission.id == AttachmentReference.reference_id, "
        "AttachmentReference.reference_type == 'manual_task_submission')",
        secondaryjoin="Attachment.id == AttachmentReference.attachment_id",
        order_by="Attachment.created_at",
        viewonly=True,
        uselist=True,
    )

    user = relationship(
        "FidesUser",
        primaryjoin="FidesUser.id == ManualTaskSubmission.submitted_by",
        viewonly=True,
    )



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any, Optional

from sqlalchemy import Column, DateTime, ForeignKey, Index, String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship
from sqlalchemy.sql import func

from fides.api.cryptography.cryptographic_util import generate_secure_random_string
from fides.api.db.base_class import Base, FidesBase
from fides.api.util.identity_verification import IdentityVerificationMixin
from fides.config import get_config

CONFIG = get_config()


if TYPE_CHECKING:
    from fides.api.models.fides_user import FidesUser

# Access links stay active for 45 days - the same as the DSR expiration. A new link is generated for each email.
# The emails are created for new DSRs which are assigned to the respondent.
ACCESS_LINK_TTL_DAYS = 45


class FidesUserRespondentEmailVerification(Base, IdentityVerificationMixin):
    """Model for handling email verification for external respondents.

    This handles two types of verification:
    1. Access links - long-lived (45 days) for initial access
    2. Verification codes - short-lived (1 hour) for actual verification

    When an email is sent to an external respondent, a new verification is created with a new access token is created.
    When a respondent clicks the link in the email, the access token is verified and a verification code is generated.
    The verification code is sent to the respondent's email address and the respondent is prompted to enter the code.
    Verification is handled by the `IdentityVerificationMixin` class.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "fides_user_respondent_email_verification"

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    created_at = Column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
        nullable=False,
    )

    user_id = Column(
        String,
        ForeignKey("fidesuser.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )

    access_token = Column(
        String, nullable=False, unique=True, index=True
    )  # Token for the access link
    access_token_expires_at = Column(DateTime(timezone=True), nullable=False)
    identity_verified_at = Column(DateTime(timezone=True), nullable=True)

    user = relationship(
        "FidesUser",
        back_populates="email_verifications",
        foreign_keys=[user_id],
    )

    __table_args__ = (
        Index(
            "ix_fides_user_respondent_email_verification_id",
            "id",
            unique=True,
        ),
    )

    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = False
    ) -> FidesUserRespondentEmailVerification:
        """
        Create a FidesUserEmailVerification record with a new access token.
        The verification code will be generated when the access link is used.
        """
        # Generate a secure token for the access link
        access_token = generate_secure_random_string(32)
        expires_at = datetime.now(timezone.utc) + timedelta(days=ACCESS_LINK_TTL_DAYS)

        verification = super().create(
            db,
            data={
                "user_id": data["user_id"],
                "access_token": access_token,
                "access_token_expires_at": expires_at,
            },
        )

        return verification

    def is_access_token_expired(self) -> bool:
        """Check if the access token has expired."""
        if not self.access_token_expires_at:
            return True

        current_time_utc = datetime.now(timezone.utc)
        return current_time_utc > self.access_token_expires_at

    def verify_access_token(self, token: str) -> bool:
        """Verify the access token and generate a verification code if valid."""
        if self.is_access_token_expired():
            return False

        return self.access_token == token

    def verify_identity(
        self,
        db: Session,
        provided_code: Optional[str] = None,
    ) -> None:
        """A method to call the internal identity verification method provided by the
        `IdentityVerificationMixin`."""
        if self.is_access_token_expired():
            raise ValueError("Access token has expired.")

        self._verify_identity(provided_code=provided_code)
        self.identity_verified_at = datetime.now(timezone.utc)
        self.save(db)



"""Add comments and comment references

Revision ID: 69ad6d844e21
Revises: 6ea2171c544f
Create Date: 2025-03-03 16:31:22.495305

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "69ad6d844e21"
down_revision = "6ea2171c544f"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "comment",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column("user_id", sa.String(), nullable=True),
        sa.Column("comment_text", sa.String(), nullable=False),
        sa.Column("comment_type", sa.String(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(["user_id"], ["fidesuser.id"], ondelete="SET NULL"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "comment_reference",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column("comment_id", sa.String(), nullable=False),
        sa.Column("reference_id", sa.String(), nullable=False),
        sa.Column("reference_type", sa.String(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(["comment_id"], ["comment.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("comment_id", "reference_id"),
    )
    # Add index on comment_reference.reference_id
    op.create_index(
        "ix_comment_reference_reference_id",
        "comment_reference",
        ["reference_id"],
    )
    # Add index on comment_reference.reference_type
    op.create_index(
        "ix_comment_reference_reference_type",
        "comment_reference",
        ["reference_type"],
    )
    # ### end Alembic commands ###


def downgrade():
    # Drop the index on comment_reference.reference_id
    op.drop_index("ix_comment_reference_reference_id", table_name="comment_reference")
    # Drop the index on comment_reference.reference_type
    op.drop_index("ix_comment_reference_reference_type", table_name="comment_reference")
    op.drop_table("comment_reference")
    op.drop_table("comment")
    # ### end Alembic commands ###



from typing import TYPE_CHECKING, Any, Optional, cast

from loguru import logger
from sqlalchemy import select
from sqlalchemy.orm import Session

from fides.api.models.fides_user import FidesUser
from fides.api.models.manual_tasks.manual_task import ManualTask, ManualTaskReference
from fides.api.models.manual_tasks.manual_task_config import ManualTaskConfig
from fides.api.schemas.manual_tasks.manual_task_schemas import (
    ManualTaskParentEntityType,
    ManualTaskReferenceType,
    ManualTaskType,
)
from fides.service.manual_tasks.manual_task_config_service import (
    ManualTaskConfigService,
)
from fides.service.manual_tasks.manual_task_instance_service import (
    ManualTaskInstanceService,
)
from fides.service.manual_tasks.utils import with_task_logging

if TYPE_CHECKING:
    from fides.api.models.manual_tasks.manual_task_instance import (
        ManualTaskInstance,
        ManualTaskSubmission,
    )


class ManualTaskError(Exception):
    """Exception raised when a manual task error occurs."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class ManualTaskService:
    def __init__(self, db: Session):
        self.db = db
        self.config_service = ManualTaskConfigService(db)
        self.instance_service = ManualTaskInstanceService(db)

    def get_task(
        self,
        task_id: Optional[str] = None,
        parent_entity_id: Optional[str] = None,
        parent_entity_type: Optional[ManualTaskParentEntityType] = None,
        task_type: Optional[ManualTaskType] = None,
    ) -> ManualTask:
        """Get the manual task using provided filters.

        This is a flexible lookup method that can find tasks based on various filters.
        It's normal for this method to return None if no task matches the given filters.

        Args:
            task_id: The task ID
            parent_entity_id: The parent entity ID
            parent_entity_type: The parent entity type
            task_type: The task type

        Returns:
            ManualTask: The matching task
        """
        if not any([task_id, parent_entity_id, parent_entity_type, task_type]):
            logger.debug("No filters provided to get_task")
            raise ManualTaskError("No filters provided to get_task")

        # Build filter conditions and a human-readable description
        filters = []
        filter_desc = []
        if task_id:
            filters.append(ManualTask.id == task_id)
            filter_desc.append(f"task_id={task_id}")
        if parent_entity_id:
            filters.append(ManualTask.parent_entity_id == parent_entity_id)
            filter_desc.append(f"parent_entity_id={parent_entity_id}")
        if parent_entity_type:
            filters.append(ManualTask.parent_entity_type == parent_entity_type)
            filter_desc.append(f"parent_entity_type={parent_entity_type}")
        if task_type:
            filters.append(ManualTask.task_type == task_type)
            filter_desc.append(f"task_type={task_type}")

        # Apply all filters at once
        stmt = select(ManualTask)  # type: ignore[arg-type]
        if filters:
            stmt = stmt.where(*filters)

        task = self.db.execute(stmt).scalar_one_or_none()
        if task is None:
            logger.debug(f"No task found with filters: {filter_desc}")
            raise ManualTaskError(f"No task found with filters: {filter_desc}")
        return task

    @with_task_logging("Provided user IDs verified")
    def _non_existent_users(
        self, non_existent_user_ids: list[str], *, task_id: str
    ) -> None:
        """This is a helper function to raise an error if users do not exist.
        If non_existent_user_ids is empty, this function will instead create
        a success log entry.

        Args:
            non_existent_user_ids: List of non-existent user IDs
            task_id: The task ID

        Returns:
            None
        """
        if len(non_existent_user_ids) > 0:
            raise ManualTaskError(
                f"User(s) {sorted(list(non_existent_user_ids))} do not exist"
            )

    def _create_log_data(self, task_id: str, details: dict[str, Any]) -> dict[str, Any]:
        """Create standard log data structure.

        Args:
            task_id: The task ID
            details: The log details

        Returns:
            dict: The log data structure
        """
        return {
            "task_id": task_id,
            "details": details,
        }

    def _handle_user_errors(
        self,
        non_existent_users: list[str],
        task_id: str,
        details: dict[str, Any],
        success_count: int,
        error_key: str,
    ) -> None:
        """Handle errors for non-existent users.

        Args:
            non_existent_users: List of non-existent user IDs
            task_id: The task ID
            details: The log details to update
            success_count: Number of successful operations
            error_key: Key to use for error details

        Raises:
            ManualTaskError: If no successful operations and users don't exist
        """
        try:
            self._non_existent_users(non_existent_users, task_id=task_id)
        except ManualTaskError as e:
            details[error_key] = sorted(non_existent_users)
            if success_count == 0:
                raise e

    def _get_existing_users(self, user_ids: list[str]) -> set[str]:
        """Get set of existing user IDs from the provided list.

        Args:
            user_ids: List of user IDs to check

        Returns:
            set: Set of existing user IDs
        """
        return set(
            user.id
            for user in self.db.query(FidesUser)
            .filter(FidesUser.id.in_(user_ids))
            .all()
        )

    def _handle_user_operation(
        self,
        task_id: str,
        user_ids: list[str],
        operation_type: str,
        current_users: Optional[set[str]] = None,
    ) -> tuple[set[str], dict[str, Any]]:
        """Handle user assignment/unassignment operations.

        Args:
            task_id: The task ID
            user_ids: List of user IDs to process
            operation_type: Type of operation ('assign' or 'unassign')
            current_users: Optional set of current users

        Returns:
            tuple: (processed_users, log_data)
        """
        if not (user_ids := list(set(user_ids))):
            raise ManualTaskError(f"User ID is required for {operation_type}ment")

        existing_users = set(
            u.id
            for u in self.db.query(FidesUser).filter(FidesUser.id.in_(user_ids)).all()
        )
        processed_users = (
            existing_users - (current_users or set())
            if operation_type == "assign"
            else existing_users
        )
        details = {f"{operation_type}ed_users": sorted(list(processed_users))}

        if non_existing := list(set(user_ids) - existing_users):
            try:
                self._non_existent_users(non_existing, task_id=task_id)
            except ManualTaskError as e:
                details[f"user_ids_not_{operation_type}ed"] = sorted(non_existing)
                if not processed_users:
                    raise e

        return processed_users, {"task_id": task_id, "details": details}

    @with_task_logging("Assign users to task")
    def assign_users_to_task(
        self, task_id: str, user_ids: list[str]
    ) -> tuple[ManualTask, dict[str, Any]]:
        """Assign users to a task.

        Args:
            task_id: The task ID
            user_ids: List of user IDs to assign

        Returns:
            Tuple containing the task and log data, the log data is
            captured by the with_task_logging decorator. and the task is
            returned to the caller.
        """
        task = self.get_task(task_id=task_id)
        users_to_assign, log_data = self._handle_user_operation(
            task_id, user_ids, "assign", set(task.assigned_users)
        )

        if users_to_assign:
            self.db.bulk_insert_mappings(
                ManualTaskReference,
                [
                    {
                        "task_id": task.id,
                        "reference_id": user_id,
                        "reference_type": ManualTaskReferenceType.assigned_user,
                    }
                    for user_id in users_to_assign
                ],
            )
            self.db.flush()
            self.db.refresh(task)

        return task, log_data

    @with_task_logging("Unassign users from task")
    def unassign_users_from_task(
        self, task_id: str, user_ids: list[str]
    ) -> tuple[ManualTask, dict[str, Any]]:
        task = self.get_task(task_id=task_id)
        refs_to_unassign = (
            self.db.query(ManualTaskReference)
            .filter(
                ManualTaskReference.task_id == task.id,
                ManualTaskReference.reference_type
                == ManualTaskReferenceType.assigned_user,
                ManualTaskReference.reference_id.in_(user_ids),
            )
            .all()
        )

        if refs_to_unassign:
            self.db.query(ManualTaskReference).filter(
                ManualTaskReference.id.in_([ref.id for ref in refs_to_unassign])
            ).delete(synchronize_session=False)
            self.db.flush()
            self.db.refresh(task)

        _, log_data = self._handle_user_operation(task_id, user_ids, "unassign")
        return task, log_data

    def create_config(
        self, config_type: str, fields: list[dict], task_id: str
    ) -> ManualTaskConfig:
        """Create a new config for a task.

        Args:
            config_type: The config type
            fields: The fields for the config
            task_id: The task ID

        Returns:
            ManualTaskConfig: The new config
        """
        task = self.get_task(task_id=task_id)
        config = self.config_service.create_new_version(task, config_type, fields)
        return cast(ManualTaskConfig, config)

    def delete_config(self, config: ManualTaskConfig, task_id: str) -> None:
        """Delete this configuration.
        Args:
            config: The config to delete
            task_id: The task ID
        Raises:
            ManualTaskConfigError: If there are active instances using this configuration
            ManualTaskError: If the task does not exist

        Returns:
            dict[str, Any]: The log details - intercepted by the `with_task_logging` decorator.
        """
        task = self.get_task(task_id=task_id)

        # Delete the configuration
        config_id = config.id
        self.config_service.delete_config(task, config_id)

    def create_instance(
        self, task_id: str, config_id: str, entity_id: str, entity_type: str
    ) -> "ManualTaskInstance":
        """Create a new instance for a task.

        Args:
            task_id: The task ID
            config_id: The config ID
            entity_id: The entity ID
            entity_type: The entity type

        Returns:
            ManualTaskInstance: The new instance
        """
        self.get_task(task_id=task_id)
        instance = self.instance_service.create_instance(
            task_id, config_id, entity_id, entity_type
        )
        return cast("ManualTaskInstance", instance)

    def create_submission(
        self, instance_id: str, field_id: str, data: dict[str, Any]
    ) -> "ManualTaskSubmission":
        """Create a new submission for a task.

        Args:
            instance_id: The instance ID
            field_id: The field ID
            data: The data for the submission

        Returns:
            ManualTaskSubmission: The new submission
        """
        submission = self.instance_service.create_submission(
            instance_id, field_id, data
        )
        return cast("ManualTaskSubmission", submission)



from pydantic import Field, model_validator
from pydantic_settings import SettingsConfigDict

from .fides_settings import FidesSettings


class ConsentSettings(FidesSettings):
    """Configuration settings for Consent."""

    tcf_enabled: bool = Field(
        default=False, description="Toggle whether TCF is enabled."
    )
    ac_enabled: bool = Field(
        default=False, description="Toggle whether Google AC Mode is enabled."
    )

    override_vendor_purposes: bool = Field(
        default=False,
        description="Whether or not vendor purposes can be globally overridden.",
    )
    model_config = SettingsConfigDict(env_prefix="FIDES__CONSENT__")

    @model_validator(mode="after")
    def validate_fields(self) -> "ConsentSettings":
        """AC mode only works if TCF mode is also enabled"""
        tcf_mode = self.tcf_enabled
        ac_mode = self.ac_enabled
        override_vendor_purposes = self.override_vendor_purposes

        if ac_mode and not tcf_mode:
            raise ValueError("AC cannot be enabled unless TCF mode is also enabled.")

        if override_vendor_purposes and not tcf_mode:
            raise ValueError(
                "Override vendor purposes cannot be true unless TCF mode is also enabled."
            )

        return self



from typing import Dict, List, Literal, Optional

from fastapi import Depends, HTTPException, status
from loguru import logger
from pydantic import BaseModel
from redis.exceptions import ResponseError
from sqlalchemy.orm import Session

import fides
from fides.api.api.deps import get_db
from fides.api.common_exceptions import RedisConnectionError
from fides.api.db.database import get_db_health
from fides.api.tasks import celery_app, get_worker_ids
from fides.api.util.api_router import APIRouter
from fides.api.util.cache import get_cache, get_queue_counts
from fides.api.util.logger import Pii
from fides.config import CONFIG

CacheHealth = Literal["healthy", "unhealthy", "no cache configured"]
HEALTH_ROUTER = APIRouter(tags=["Health"])


class CoreHealthCheck(BaseModel):
    """Server Healthcheck schema"""

    webserver: str
    version: str
    cache: CacheHealth


class DatabaseHealthCheck(BaseModel):
    """Database Healthcheck Schema"""

    database: str
    database_revision: Optional[str] = None


class WorkerHealthCheck(BaseModel):
    """Worker Healthcheck Schema"""

    workers_enabled: bool
    workers: List[str]
    queue_counts: Dict[str, int]


def get_cache_health() -> str:
    """Checks if the cache is reachable"""

    if not CONFIG.redis.enabled:
        return "no cache configured"
    try:
        get_cache()
        return "healthy"
    except (RedisConnectionError, ResponseError) as e:
        logger.error("Unable to reach cache: {}", Pii(str(e)))
        return "unhealthy"


@HEALTH_ROUTER.get(
    "/health/database",
    response_model=DatabaseHealthCheck,
    responses={
        status.HTTP_200_OK: {
            "content": {
                "application/json": {
                    "example": {
                        "database": "healthy",
                    }
                }
            }
        },
        status.HTTP_503_SERVICE_UNAVAILABLE: {
            "content": {
                "application/json": {
                    "example": {
                        "detail": {
                            "database": "unhealthy",
                        }
                    }
                }
            }
        },
    },
)
async def database_health(db: Session = Depends(get_db)) -> Dict:
    """Confirm that the API is running and healthy."""
    db_health, revision = get_db_health(CONFIG.database.sync_database_uri, db=db)

    response = DatabaseHealthCheck(
        database=db_health, database_revision=revision if revision else "unknown"
    ).model_dump(mode="json")

    if db_health != "healthy":
        raise HTTPException(
            status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail=response
        )

    return response


@HEALTH_ROUTER.get(
    "/health/workers",
    response_model=WorkerHealthCheck,
    responses={
        status.HTTP_200_OK: {
            "content": {
                "application/json": {
                    "example": {
                        "workers_enabled": "True",
                        "workers": ["celery@c606808353b5"],
                    }
                }
            }
        },
        status.HTTP_503_SERVICE_UNAVAILABLE: {
            "content": {
                "application/json": {
                    "example": {
                        "detail": {
                            "workers_enabled": "True",
                            "workers": ["celery@c606808353b5"],
                        }
                    }
                }
            }
        },
    },
)
async def workers_health() -> Dict:
    """Confirm that the API is running and healthy."""
    response = WorkerHealthCheck(
        workers_enabled=False, workers=[], queue_counts={}
    ).model_dump(mode="json")

    fides_is_using_workers = not celery_app.conf["task_always_eager"]
    if fides_is_using_workers:
        response["workers_enabled"] = True
        # Figure out a way to make this faster
        response["workers"] = get_worker_ids()
        response["queue_counts"] = get_queue_counts()

    return response


@HEALTH_ROUTER.get(
    "/health",
    response_model=CoreHealthCheck,
    responses={
        status.HTTP_200_OK: {
            "content": {
                "application/json": {
                    "example": {
                        "webserver": "healthy",
                        "version": "1.0.0",
                        "cache": "healthy",
                    }
                }
            }
        },
        status.HTTP_503_SERVICE_UNAVAILABLE: {
            "content": {
                "application/json": {
                    "example": {
                        "detail": {
                            "webserver": "healthy",
                            "version": "1.0.0",
                            "cache": "unhealthy",
                        }
                    }
                }
            }
        },
    },
)
async def health() -> Dict:
    """Confirm that the API is running and healthy."""
    cache_health = get_cache_health()
    response = CoreHealthCheck(
        webserver="healthy",
        version=str(fides.__version__),
        cache=cache_health,
    ).model_dump(mode="json")

    for _, value in response.items():
        if value == "unhealthy":
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail=response
            )

    return response



from fides.api.models.manual_tasks.manual_task import ManualTask, ManualTaskReference
from fides.api.models.manual_tasks.manual_task_config import (
    ManualTaskConfig,
    ManualTaskConfigField,
)
from fides.api.models.manual_tasks.manual_task_instance import (
    ManualTaskInstance,
    ManualTaskSubmission,
)
from fides.api.models.manual_tasks.manual_task_log import ManualTaskLog

__all__ = [
    "ManualTask",
    "ManualTaskConfig",
    "ManualTaskConfigField",
    "ManualTaskInstance",
    "ManualTaskLog",
    "ManualTaskReference",
    "ManualTaskSubmission",
]



from typing import Any, Dict, Optional

from boto3 import Session
from botocore.exceptions import ClientError
from loguru import logger

from fides.api.common_exceptions import StorageUploadError
from fides.api.schemas.storage.storage import AWSAuthMethod, StorageSecrets


def get_aws_session(
    auth_method: str,
    storage_secrets: Optional[Dict[StorageSecrets, Any]],
    assume_role_arn: Optional[str] = None,
) -> Session:
    """
    Abstraction to retrieve an AWS Session using secrets.

    If an `assume_role_arn` is provided, the secrets will be used to
    assume that role and return a Session instantiated with that role.
    """
    if storage_secrets is None:
        # set to an empty dict to allow for more dynamic code downstream
        storage_secrets = {}

    stored_region_name = storage_secrets.get("region_name")  # type: ignore
    stored_assume_role_arn: Optional[str] = storage_secrets.get("assume_role_arn")  # type: ignore

    if auth_method == AWSAuthMethod.SECRET_KEYS.value:
        if not storage_secrets:
            err_msg = "Storage secrets not found for S3 storage."
            logger.warning(err_msg)
            raise StorageUploadError(err_msg)

        session = Session(
            aws_access_key_id=storage_secrets[StorageSecrets.AWS_ACCESS_KEY_ID.value],  # type: ignore
            aws_secret_access_key=storage_secrets[
                StorageSecrets.AWS_SECRET_ACCESS_KEY.value  # type: ignore
            ],
            region_name=stored_region_name,
        )
    elif auth_method == AWSAuthMethod.AUTOMATIC.value:
        session = Session(region_name=stored_region_name)
        logger.info("Successfully created automatic session")
    else:
        logger.error("AWS auth method not supported: {}", auth_method)
        raise ValueError(f"AWS auth method not supported: {auth_method}")

    # Check that credentials are valid
    sts_client = session.client("sts")
    sts_client.get_caller_identity()

    target_assume_role_arn = assume_role_arn or stored_assume_role_arn
    if target_assume_role_arn:
        try:
            return get_assumed_role_session(
                target_assume_role_arn,
                sts_client,
                stored_region_name,
            )
        except ClientError as error:
            logger.exception(
                f"Failed to assume assume role {assume_role_arn}. Error: {error.response['Error']['Message']}"
            )
            raise
    else:
        return session


def get_assumed_role_session(
    assume_role_arn: str, sts_client: Any, region_name: Optional[str] = None
) -> Session:
    response = sts_client.assume_role(
        RoleArn=assume_role_arn, RoleSessionName="FidesAssumeRoleSession"
    )
    temp_credentials = response["Credentials"]
    logger.info(f"Assumed role {assume_role_arn} and got temporary credentials.")
    return Session(
        aws_access_key_id=temp_credentials["AccessKeyId"],
        aws_secret_access_key=temp_credentials["SecretAccessKey"],
        aws_session_token=temp_credentials["SessionToken"],
        region_name=region_name,
    )


def get_s3_client(
    auth_method: str,
    storage_secrets: Optional[Dict[StorageSecrets, Any]],
    assume_role_arn: Optional[str] = None,
) -> Session:
    """
    Abstraction to retrieve an AWS S3 client using secrets.

    If an `assume_role_arn` is provided, the secrets will be used to
    assume that role and return a Session instantiated with that role.
    """
    session = get_aws_session(
        auth_method=auth_method,
        storage_secrets=storage_secrets,
        assume_role_arn=assume_role_arn,
    )
    return session.client("s3")



from typing import Any

from fideslang.models import FidesDatasetReference
from pydantic import Field
from sqlalchemy.orm import Session

from fides.api.common_exceptions import ValidationError
from fides.api.models.datasetconfig import validate_dataset_reference
from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets_email import (
    BaseEmailSchema,
)


class DynamicErasureEmailSchema(BaseEmailSchema):
    """
    Schema to validate the secrets needed for a dynamic erasure email connector.
    third_party_vendor_name and recipient_email_address must reference the same dataset
    and collection, e.g "dataset.publishers.vendor_name" and "dataset.publishers.email"
    """

    third_party_vendor_name: FidesDatasetReference = Field(
        title="Third party vendor name field",
        description="Dataset reference to the field containing the third party vendor name. Both third_party_vendor_name and recipient_email_address must reference the same dataset and collection.",
        json_schema_extra={
            "external_reference": True
        },  # metadata added so we can identify these secret schema fields as external references
    )
    recipient_email_address: FidesDatasetReference = Field(
        title="Recipient email address field",
        description="Dataset reference to the field containing the recipient email address. Both third_party_vendor_name and recipient_email_address must reference the same dataset and collection.",
        json_schema_extra={
            "external_reference": True
        },  # metadata added so we can identify these secret schema fields as external references
    )


class DynamicErasureEmailDocsSchema(DynamicErasureEmailSchema, NoValidationSchema):
    """DynamicErasureEmailDocsSchema Secrets Schema for API Docs"""


def validate_dynamic_erasure_email_dataset_references(
    db: Session,
    connection_secrets: "DynamicErasureEmailSchema",
) -> Any:
    """
    Validates that the provided connection secrets reference an existing datasetconfig collection,
    and that both third_party_vendor_name and recipient_email_address reference the same collection
    from the same dataset. If any of these conditions are not met, a ValidationError is raised.
    """
    validate_dataset_reference(db, connection_secrets.recipient_email_address)
    validate_dataset_reference(db, connection_secrets.third_party_vendor_name)

    if (
        connection_secrets.recipient_email_address.dataset
        != connection_secrets.third_party_vendor_name.dataset
    ):
        raise ValidationError(
            "Recipient email address and third party vendor name must reference the same dataset"
        )

    email_collection_name = connection_secrets.recipient_email_address.field.split(".")[
        0
    ]
    vendor_collection_name = connection_secrets.third_party_vendor_name.field.split(
        "."
    )[0]

    if email_collection_name != vendor_collection_name:
        raise ValidationError(
            "Recipient email address and third party vendor name must reference the same collection"
        )



import secrets
from typing import Optional, Union

from fides.api.cryptography.cryptographic_util import bytes_to_b64_str
from fides.api.util.cache import get_cache, get_encryption_cache_key
from fides.api.util.encryption.aes_gcm_encryption_scheme import (
    encrypt_to_bytes_verify_secrets_length,
)
from fides.config import CONFIG


def encrypt_access_request_results(data: Union[str, bytes], request_id: str) -> str:
    """Encrypt data with encryption key if provided, otherwise return unencrypted data.

    Args:
        data: The data to encrypt
        request_id: The ID of the privacy request for encryption key lookup

    Returns:
        str: The encrypted data as a string
    """
    cache = get_cache()
    encryption_cache_key = get_encryption_cache_key(
        privacy_request_id=request_id,
        encryption_attr="key",
    )
    if isinstance(data, bytes):
        data = data.decode(CONFIG.security.encoding)

    encryption_key: Optional[str] = cache.get(encryption_cache_key)
    if not encryption_key:
        return data

    bytes_encryption_key: bytes = encryption_key.encode(
        encoding=CONFIG.security.encoding
    )
    nonce: bytes = secrets.token_bytes(CONFIG.security.aes_gcm_nonce_length)
    # b64encode the entire nonce and the encrypted message together
    return bytes_to_b64_str(
        nonce
        + encrypt_to_bytes_verify_secrets_length(data, bytes_encryption_key, nonce)
    )



import re
from typing import Any, Dict, List, Optional, Tuple

from fides.api.graph.config import Field
from fides.api.models.policy import Policy
from fides.api.service.connectors.query_configs.query_config import SQLLikeQueryConfig

ScyllaDBStatement = Tuple[str, Dict[str, Any]]
"""
A ScyllaDB statement is a tuple whose first component is the query string
and whose second component is a dict of parameters to pass in to the query.

E.g
('SELECT name FROM users WHERE user.id = %(user_id)s', { "user_id": 123 })
"""


class ScyllaDBQueryConfig(SQLLikeQueryConfig[ScyllaDBStatement]):
    def query_to_str(
        self, t: ScyllaDBStatement, input_data: Dict[str, List[Any]]
    ) -> str:
        """String representation of a query for logging/dry-run"""

        def transform_param(param: Any) -> str:
            if isinstance(param, str):
                # Is param is already a string, then we need to add single quotes
                return f"'{param}'"
            return str(param)

        query_str, query_params = t

        for param_name, param_value in query_params.items():
            # We look up the places where we have the param name as the placeholder for the value
            # And replace each instance with the actual value
            query_str = re.sub(
                rf"%\({param_name}\)s", f"{transform_param(param_value)}", query_str
            )

        return query_str

    def dry_run_query(self) -> Optional[str]:
        """Returns a text representation of the query."""
        query_data = self.display_query_data()
        query_statement = self.generate_query(query_data, None)
        if not query_statement:
            return None

        return self.query_to_str(query_statement, query_data)

    def format_clause_for_query(
        self, string_path: str, operator: str, operand: str
    ) -> str:
        """Returns clauses in a format they can be added into ScyllaDB CQL queries."""
        if operator == "IN":
            return f"{string_path} IN ({operand})"
        return f"{string_path} {operator} %({operand})s"

    def get_formatted_query_string(
        self,
        field_list: str,
        clauses: List[str],
    ) -> str:
        """Returns an SQL query string."""
        return f"SELECT {field_list} FROM {self.node.collection.name} WHERE {' OR '.join(clauses)} ALLOW FILTERING;"

    def generate_query(
        self,
        input_data: Dict[str, List[Any]],
        policy: Optional[Policy] = None,
    ) -> Optional[ScyllaDBStatement]:
        return self.generate_query_without_tuples(input_data, policy)

    def format_key_map_for_update_stmt(self, param_map: Dict[str, Any]) -> List[str]:
        """Adds the appropriate formatting for update statements in this datastore."""
        return [f"{k} = %({v})s" for k, v in sorted(param_map.items())]

    def get_update_clauses(
        self,
        update_value_map: Dict[str, Any],
        where_clause_fields: Dict[str, Field],
    ) -> List[str]:
        """Returns a list of update clauses for the update statement.

        Omits primary key fields from updates since ScyllaDB prohibits
        updating primary key fields.
        """

        return self.format_key_map_for_update_stmt(
            {
                key: value
                for key, value in update_value_map.items()
                if key not in where_clause_fields
            }
        )

    def format_query_data_name(self, query_data_name: str) -> str:
        return f"%({query_data_name})s"

    def format_query_stmt(
        self, query_str: str, update_value_map: Dict[str, Any]
    ) -> ScyllaDBStatement:
        return (query_str, update_value_map)



from sqlalchemy import Column, ForeignKey, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict

from fides.api.db.base_class import Base  # type: ignore[attr-defined]
from fides.api.models.fides_user import FidesUser


class CustomReport(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "plus_custom_report"

    name = Column(String, unique=True)
    type = Column(String, nullable=False)
    created_by = Column(
        String,
        ForeignKey(FidesUser.id_field_path, ondelete="SET NULL"),
        nullable=True,
    )
    config = Column(MutableDict.as_mutable(JSONB))
